# Copyright (C) 2016-2021 J.F.Dockes
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the
#   Free Software Foundation, Inc.,
#   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
#
"""
Shared code for the upmpdcli mediaserver plugins.

 - Uses the interface for the entity objects (track, album...)
   concretely defined in upmplgmodels.py, but duck-typed.
 - Defines and uses the format for the permanent URLs

The module implements utility functions for translating to/from what
our parent expects or sends on the pipe.
"""

import sys
import os
import pwd
import errno
import inspect
import posixpath
import re
import unicodedata

import conftree

# This is only used for log messages
_idprefix = "0$UNKNOWN"


def setidprefix(idprefix):
    global _idprefix
    _idprefix = idprefix


def direntry(
        id,
        pid,
        title,
        arturi=None,
        artist=None,
        upnpclass=None,
        searchable="1",
        date=None,
        description=None,
        genre=None,
):
    """Create container entry in format expected by parent"""
    # uplog(f"rcldirentry: id {id} pid {pid} tt {title} date {date} clss {upnpclass} artist {artist} arturi {arturi}")
    ret = {"id": id, "pid": pid, "tt": title, "tp": "ct", "searchable": searchable}
    if arturi:
        ret["upnp:albumArtURI"] = arturi
    if artist:
        ret["upnp:artist"] = artist
    if date:
        ret["dc:date"] = date
    if upnpclass:
        ret["upnp:class"] = upnpclass
    else:
        ret["upnp:class"] = "object.container.storageFolder"
    if description:
        ret["dc:description"] = description
    if genre:
        ret["upnp:genre"] = genre
    return ret


# Generate an permanent URL in the form which can be used trackid_from_urlpath
def url_from_trackid(httphp, pathprefix, trackid):
    return f"http://{httphp}" + posixpath.join(pathprefix, f"track/version/1/trackId/{trackid}")


# Extract the service (qobuz,tidal...) trackid from one of our special form permanent URLs.
#
# This typically gets called from the plugin trackuri() method, which is used for translating from
# our permanent URLs to a service temporary URL when the renderer tries to fetch the stream.
def trackid_from_urlpath(pathprefix, a):
    """
    Extract track id from a permanent URL path part.

    This supposes that the input URL has the format produced by urlpath_from_trackid()
    (called from, e.g., trackentries()): <pathprefix>/track/version/1/trackId/<trackid>

    Args:
        pathprefix (str): our configured path prefix (e.g. /qobuz/)
        a (dict): the argument dict out of cmdtalk with a 'path' key
    Returns:
        str: the track Id.
    """

    if "path" not in a:
        raise Exception("trackuri: no 'path' in args")
    path = a["path"]

    # pathprefix + 'track/version/1/trackId/trackid
    exp = posixpath.join(pathprefix, r"track/version/1/trackId/(.+)$")
    m = re.match(exp, path)
    if m is None:
        exp_old = posixpath.join(pathprefix, r"track\?version=1&trackId=(.+)$")
        m = re.match(exp_old, path)
    if m is None:
        raise Exception(f"trackuri: path [{path}] does not match [{exp}]")
    trackid = m.group(1)
    return trackid


# NPUPNP web server document root if set. This does not come directly from the config, but uses a
# specific environment variable (upmpdcli does some processing on the configuration value).
def getUpnpWebDocRoot(servicename):
    try:
        d = os.environ["UPMPD_UPNPDOCROOT"]
        dp = os.path.join(d, servicename)
        if not os.path.exists(dp):
            os.makedirs(dp)
        # returning /.../cachedir/www not /.../cachedir/www/pluginname 
        return d
    except Exception as ex:
        #uplog(f"NO UPNPWEBDOCROOT: {ex}")
        return ""
    

# Generic option retrieval: get the value from the configuration file or the environment.
_g_upconfig = None
def getOptionValue(nm, dflt=None):
    global _g_upconfig
    if _g_upconfig is None:
        members = dict(inspect.getmembers(conftree.ConfSimple.__init__.__code__))
        var_names = members["co_varnames"]
        if "casesensitive" in var_names:
            _g_upconfig = conftree.ConfSimple(os.environ["UPMPD_CONFIG"], casesensitive=False)
        else:
            _g_upconfig = conftree.ConfSimple(os.environ["UPMPD_CONFIG"])
    value = _g_upconfig.get(nm)
    if value is None:
        envar = "UPMPD_" + nm.upper()
        try:
            value = os.environ[envar]
        except:
            value = dflt

    if dflt is not None and type(dflt) != type(""):
        # Try to convert the value to the type of the provided default
        try:
            # Bool("0") is true, so special-case the bool type
            if type(dflt) == type(True):
                value = conftree.valToBool(value)
            else:
                t = type(dflt)
                value = t(value)
        except Exception as ex:
            uplog(f"Config: type conversion failed for {nm}")

    return value


def getConfigObject():
    if _g_upconfig is None:
        getOptionValue("somebogusvalue")
    return _g_upconfig


# Get user and password from service, from the main configuration
# file, or possibly from the ohcredentials scratchpad. In both files,
# the entries are like:
#    qobuzuser=xxx
#    qobuzpass=yyy
def getserviceuserpass(servicename):
    username = getOptionValue(servicename + "user")
    password = getOptionValue(servicename + "pass")
    if not username or not password:
        credsfile = os.path.join(getcachedir(""), "ohcreds", "screds")
        uplog("Retrieving user/pass from %s" % credsfile)
        altconf = conftree.ConfSimple(credsfile)
        username = altconf.get(servicename + "user")
        password = altconf.get(servicename + "pass")
        if not username or not password:
            uplog("Failed retrieving user/pass")
    return username, password


_loglevel = 3


def uplog(s, level=3):
    if level > _loglevel:
        return
    if not type(s) == type(b""):
        s = ("%s: %s" % (_idprefix, s)).encode("utf-8")
        sys.stderr.buffer.write(s + b"\n")
    sys.stderr.flush()


def getcachedir(servicename, forcedpath=None):
    if forcedpath:
        cachedir = forcedpath
    else:
        cachedir = getOptionValue("cachedir")
        if not cachedir:
            me = pwd.getpwuid(os.getuid()).pw_name
            uplog("getcachedir: me: %s" % me)
            if me == "upmpdcli":
                cachedir = "/var/cache/upmpdcli/"
            else:
                cachedir = os.path.expanduser("~/.cache/upmpdcli/")
        cachedir = os.path.join(cachedir, servicename)
    try:
        os.makedirs(cachedir)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(cachedir):
            pass
        else:
            raise
    return cachedir


default_mime = "audio/mpeg"
default_samplerate = "44100"
default_bits = "16"
default_channels = "2"


# Default values for when we don't know otherwise. This is mostly for display purposes anyway, the
# renderers always check and process actual the audio formats from the data
def setMimeAndSamplerate(m, s, bits=None, channels=None):
    global default_mime, default_samplerate, default_bits, default_channels
    default_mime = m
    default_samplerate = s
    if bits:
        default_bits = str(bits)
    if channels:
        default_channels = str(channels)


# Translate an upmplgmodels Tracks array into output expected by plgwithslave. The URLs are
# constructed for future redirection/proxying by our HTTP server.
# optionally, using argument generate_track_nums set to True, tracknumbers can be generated by
# counting the presented elements, so that the control points which decide to apply track sorting
# will still present the tracks in the correct order in the case of playlists and, most importantly,
# multi-disc albums This is afaik to the benefit of kodi mostly
def trackentries(httphp, pathprefix, objid, tracks, generate_track_nums=False):
    """
    Transform a list of Track objects to the format expected by the parent

    Args:
        objid (str):  objid for the browsed object (the parent container)
        tracks is the array of Track objects to be translated
        tracks: a list of Track objects.
        generate_track_nums: boolean values, if set, the tracknumber is generated

    Returns:
        A list of dicts, each representing an UPnP item, with the
        keys as expected in the plgwithslave.cxx resultToEntries() function.

        The permanent URIs, are of the following form, based on the
        configured host:port and pathprefix arguments and track Id:
            http://host:port/pathprefix/track/version/1/trackId/<trackid>
        We now call url_from_trackid() to generate them
    """
    entries = []
    track_counter: int = 0
    for track in tracks:
        if not track.available:
            if 1:
                uplog("NOT AVAILABLE")
                try:
                    uplog("%s by %s" % (track.name, track.artist.name))
                except:
                    pass
            continue
        track_counter += 1
        li = {}
        li["pid"] = objid
        li["id"] = objid + "$" + "%s" % track.id
        li["tt"] = track.name
        li["uri"] = url_from_trackid(httphp, pathprefix, track.id)
        li["tp"] = "it"
        image = getattr(track, "image", None)
        if image:
            li["upnp:albumArtURI"] = image
        if track.album:
            li["upnp:album"] = track.album.name
            if not track.image and track.album.image:
                li["upnp:albumArtURI"] = track.album.image
            if track.album.release_date:
                li["dc:date"] = track.album.release_date
            # Do we really want to do this ? This would currently be the only way to display the,
            # e.gh. Qobuz album description in upplay (because the description is not set in album
            # lists used by the dir browser, only when querying a specific album), but it does not
            # seem quite right to set it on each track. Longer discussion in notes.
            # if not 'dc:description' in li and track.album.description:
            # uplog(f"trackentries: setting dc:description on track: {track.album.description}")
            # li['dc:description'] = track.album.description
        # if generate_track_nums, we use the counter instead of the metadata
        li["upnp:originalTrackNumber"] = (
            str(track_counter) if generate_track_nums else str(track.track_num)
        )
        li["upnp:artist"] = track.artist.name
        li["dc:title"] = track.name
        # if generate_track_nums, we can skip the discnumber
        if not generate_track_nums:
            li["discnumber"] = str(track.disc_num)
        li["upnp:class"] = track.upnpclass

        li["duration"] = str(track.duration)
        li["res:mime"] = track.mime if track.mime else default_mime
        li["res:samplefreq"] = track.samplefreq if track.samplefreq else default_samplerate
        li["res:bitsPerSample"] = track.bitdepth if track.bitdepth else default_bits
        li["res:channels"] = track.channels if track.channels else default_channels

        entries.append(li)
    return entries

def strip_accents(s):
   return ''.join(c for c in unicodedata.normalize('NFD', s)
                  if unicodedata.category(c) != 'Mn')



#!/usr/bin/env python3

'''
Dynamic art for Dutch Public Broadcasting Service (NPO) FunX internet radio stations:

# NPO FUNX

[radio NPO FunX]
url = http://icecast.omroep.nl/funx-bb-mp3			
artUrl = 
artScript = npo-funx-dynamic-art.py funx broadcast-info

[radio NPO FunX Dance]
url = http://icecast.omroep.nl/funx-dance-bb-mp3
artUrl = 
artScript = npo-funx-dynamic-art.py dance

[radio NPO Funx Slow Jamz]
url = http://icecast.omroep.nl/funx-slowjamz-bb-mp3
artUrl= 
artScript = npo-funx-dynamic-art.py slow-jamz

[radio NPO FunX Arab]
url = http://icecast.omroep.nl/funx-arab-bb-mp3
artUrl = 
artScript = npo-funx-dynamic-art.py amsterdam-arab

[radio NPO Funx Hip Hop]
url = http://icecast.omroep.nl/funx-hiphop-bb-mp3
artUrl = 
artScript = npo-funx-dynamic-art.py denhaag-hiphop

[radio NPO FunX Reggae]
url = http://icecast.omroep.nl/funx-reggae-bb-mp3
artUrl = 
artScript = npo-funx-dynamic-art.py rotterdam-afro

[radio NPO FunX Latin]
url = http://icecast.omroep.nl/funx-latin-bb-mp3
artUrl = 
artScript = npo-funx-dynamic-art.py utrecht-latin

[radio NPO FunX Amsterdam]
url = http://icecast.omroep.nl/funx-amsterdam-bb-mp3
artUrl = 
artScript = npo-funx-dynamic-art.py amsterdam broadcast-info

[radio NPO FunX Rotterdam]
url = http://icecast.omroep.nl/funx-rotterdam-bb-mp3
artUrl = 
artScript = npo-funx-dynamic-art.py rotterdam broadcast-info

[radio NPO FunX Den Haag]
url = http://icecast.omroep.nl/funx-denhaag-bb-mp3
artUrl = 
artScript = npo-funx-dynamic-art.py den-haag broadcast-info

[radio NPO FunX Utrecht]
url = http://icecast.omroep.nl/funx-utrecht-bb-mp3
artUrl = 
artScript = npo-funx-dynamic-art.py utrecht broadcast-info

'''


import json
import requests
import sys
import time
from datetime import datetime


def elemExists(obj, elem):
  if elem in obj:
    return True
  return False

def mkStr(elemStr):
  if elemStr:
    return str(elemStr)
  return ''


delay          = 0
offset         = 2
validtrack     = False
validtrackart  = False
validbroadcast = False


if len(sys.argv) > 1:
  if str(sys.argv[1]) == 'funx':
    channel  = ''
  else:
    channel  =  mkStr(sys.argv[1])
  time.sleep(delay)
  url_tracks = 'https://www.funx.nl/api/tracks/' + channel

# First, check if there's valid track art, if not check for valid broadcast art
  try:
    r2 = requests.get(url_tracks)
    r2.raise_for_status()
  except:
    pass
  else:
    try:
      newjsd_tracks = r2.json()
    except:
      pass
    else:
      if elemExists(newjsd_tracks, 'data'):
        now = int(time.time())
        # Assumption: tracks are listed in descending order (start time)
        for track in newjsd_tracks['data']:
          if elemExists(track, 'startdatetime'):
            trackstart = int(datetime.timestamp(datetime.strptime(track['startdatetime'], '%Y-%m-%dT%H:%M:%S')))
            trackend   = int(datetime.timestamp(datetime.strptime(track['stopdatetime'],   '%Y-%m-%dT%H:%M:%S')))
            if trackstart < now + offset and (trackend > now + offset - 1 or trackstart == trackend):
              validtrack = True
              break
            if trackstart < now + offset and trackend < now + offset:
              break
        if validtrack:
          artUrl = mkStr(track['image_url']).replace('{format}', 'w_1080,h_1080')
          if artUrl:
            validtrackart = True
            print(artUrl)
		  
  if not validtrackart and len(sys.argv) > 2:

    if str(sys.argv[2]) == 'broadcast-info':
      url_broadcasts = 'https://www.funx.nl/api/broadcasts/' + channel

    try:
      r1 = requests.get(url_broadcasts)
      r1.raise_for_status()
    except:
      pass
    else:
      try:
        newjsd_broadcasts = r1.json()
      except:
        pass
      else:
        if elemExists(newjsd_broadcasts, 'data'):
          # Assumption: broadcasts are listed in descending order (start time)
          for broadcast in newjsd_broadcasts['data']:
            if elemExists(broadcast, 'startdatetime'):
              trackstart = int(datetime.timestamp(datetime.strptime(broadcast['startdatetime'], '%Y-%m-%dT%H:%M:%S')))
              trackend   = int(datetime.timestamp(datetime.strptime(broadcast['stopdatetime'],   '%Y-%m-%dT%H:%M:%S')))
              if trackstart < now + offset and (trackend > now + offset - 1 or trackstart == trackend):
                validbroadcast = True
                break
              if trackstart < now + offset and trackend < now + offset:
                break
          if validbroadcast:
            artUrl = mkStr(broadcast['image_url']).replace('{format}', 'w_1080,h_1080')
            if artUrl:
              print(artUrl)


# Copyright (C) 2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import constants


class ReleaseDate:

    def __init__(self, obj_dict: dict[str, int]):
        self.__dict: dict[str, int] = {}
        k: constants.DictKey
        for k in [constants.DictKey.YEAR, constants.DictKey.MONTH, constants.DictKey.DAY]:
            if k.value in obj_dict:
                self.__dict[k.value] = obj_dict[k.value]

    def __or_none(self, dict_key: constants.DictKey) -> str | None:
        return self.__dict[dict_key.value] if dict_key.value in self.__dict else None

    def __or_none_as_int(self, dict_key: constants.DictKey) -> str | None:
        or_none: str | None = self.__or_none(dict_key=dict_key)
        if or_none is None:
            return None
        try:
            return int(or_none)
        except Exception:
            # maybe log something?
            return None

    @property
    def year(self) -> int:
        return self.__or_none_as_int(dict_key=constants.DictKey.YEAR)

    @property
    def month(self) -> int:
        return self.__or_none_as_int(dict_key=constants.DictKey.MONTH)

    @property
    def day(self) -> int:
        return self.__or_none_as_int(dict_key=constants.DictKey.DAY)



# Copyright (C) 2023 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import tidalapi

from enum import Enum


class SearchType(Enum):
    ALBUM = 0, "album", "albums", tidalapi.album.Album
    ARTIST = 1, "artist", "artists", tidalapi.artist.Artist
    TRACK = 2, "title", "tracks", tidalapi.media.Track

    def __init__(
            self,
            num: int,
            element_name: str,
            dict_entry: str,
            model: type):
        self.num: int = num
        self.element_name: str = element_name
        self.dict_entry: str = dict_entry
        self.model: type = model

    def get_name(self):
        return self.element_name

    def get_dict_entry(self) -> str:
        return self.dict_entry

    def get_model(self) -> type:
        return self.model


# duplicate check
name_checker_set: set[str] = set()
id_checker_set: set[int] = set()
for v in SearchType:
    if v.get_name() in name_checker_set:
        raise Exception(f"Duplicated name [{v.get_name()}]")
    if v.value[0] in id_checker_set:
        raise Exception(f"Duplicated id [{v.value[0]}]")
    name_checker_set.add(v.get_name())
    id_checker_set.add(v.value[0])



# Copyright (C) 2023 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class TrackAdapter:

    def get_id(self) -> str:
        pass

    def get_name(self) -> str:
        pass

    def get_duration(self) -> int:
        pass

    def get_track_num(self) -> int:
        pass

    def get_volume_num(self) -> int:
        pass

    def get_album_num_volumes(self) -> int:
        pass

    def get_album_track_count(self) -> int:
        pass

    def get_album_id(self) -> str:
        pass

    def get_album_name(self) -> str:
        pass

    def get_album_artist_name(self) -> str:
        pass

    def get_audio_quality(self) -> str:
        pass

    def get_image_url(self) -> str:
        pass

    def explicit(self) -> bool:
        pass

    def get_artist_name(self) -> str:
        pass

    def get_bit_depth(self) -> int:
        pass

    def get_sample_rate(self) -> int:
        pass



#!/usr/bin/env python3

'''
Dynamic art for Dutch Public Broadcasting Service (NPO) internet radio stations:

# NPO

[radio NPO Radio 1]
url = http://icecast.omroep.nl/radio1-bb-mp3
artUrl = 
artScript = npo-dynamic-art.py nporadio1

[radio NPO Radio 2]
url = http://icecast.omroep.nl/radio2-bb-mp3
artUrl = 
artScript = npo-dynamic-art.py nporadio2

[radio NPO 3FM]
url = http://icecast.omroep.nl/3fm-bb-mp3
artUrl = 
artScript = npo-dynamic-art.py npo3fm

[radio NPO 3FM Alternative]
url = http://icecast.omroep.nl/3fm-alternative-mp3
artUrl = 
artScript = npo-dynamic-art.py npo3fm alternative
 
[radio NPO 3FM KXradio]
url = http://icecast.omroep.nl/3fm-serioustalent-mp3
artUrl = 
artScript = npo-dynamic-art.py npo3fm kx

[radio NPO Radio 4]
url = http://icecast.omroep.nl/radio4-bb-mp3
artUrl = 
artScript = npo-dynamic-art.py nporadio4

[radio NPO Radio 4 Concerten]
url = http://icecast.omroep.nl/radio4-eigentijds-mp3
artUrl = 
artScript = npo-dynamic-art.py nporadio4 concerten

[radio NPO Radio 5]
url = http://icecast.omroep.nl/radio5-bb-mp3
artUrl = 
artScript = npo-dynamic-art.py nporadio5

[radio NPO Soul & Jazz]
url = http://icecast.omroep.nl/radio6-bb-mp3
artUrl = 
artScript = npo-dynamic-art.py nporadio2 soulenjazz

[radio NPO Sterren NL]
url = http://icecast.omroep.nl/radio2-sterrennl-mp3
artUrl = 
artScript = npo-dynamic-art.py nporadio5 sterrennl

'''


import json
import requests
import sys
import time
from datetime import datetime


def elemExists(obj, elem):
  if elem in obj:
    return True
  return False

def mkStr(elemT):
  if elemT:
    return str(elemT)
  return ''


delay          = 0
offset         = 2
validtrack     = False
validtrackart  = False
validbroadcast = False


if len(sys.argv) == 2:
  time.sleep(delay)
  url_tracks     = 'https://www.' + str(sys.argv[1]) + '.nl/api//tracks'
elif len(sys.argv) > 2:
  time.sleep(delay)
  url_tracks     = 'https://www.' + str(sys.argv[1]) + '.nl/api/' + str(sys.argv[2]) + '/tracks'

if len(sys.argv) > 1:
  # First, check if there's valid track art, if not check for valid broadcast art
  try:
    r2 = requests.get(url_tracks)
    r2.raise_for_status()
  except:
    pass
  else:
    try:
      newjsd_tracks = r2.json()
    except:
      pass
    else:
      if elemExists(newjsd_tracks, 'data'):
        now = int(time.time())
        # Assumption: tracks are listed in descending order (start time)
        for track in newjsd_tracks['data']:
          if elemExists(track, 'startdatetime'):
            trackstart = int(datetime.timestamp(datetime.strptime(track['startdatetime'], '%Y-%m-%dT%H:%M:%S')))
            trackend   = int(datetime.timestamp(datetime.strptime(track['stopdatetime'],   '%Y-%m-%dT%H:%M:%S')))
            if trackstart < now + offset and (trackend > now + offset - 1 or trackstart == trackend):
              validtrack = True
              break
            if trackstart < now + offset and trackend < now + offset:
              break
        if validtrack:
          artUrl = mkStr(track['image_url']).replace('{format}', 'w_1080,h_1080')
          if artUrl:
            validtrackart = True
            print(artUrl)

  if not validtrackart:

    if len(sys.argv) == 2:
      url_broadcasts = 'https://www.' + str(sys.argv[1]) + '.nl/api/broadcasts'
    elif len(sys.argv) > 2:
      url_broadcasts = 'https://www.' + str(sys.argv[1]) + '.nl/api/' + str(sys.argv[2]) + '/broadcasts'

    try:
      r1 = requests.get(url_broadcasts)
      r1.raise_for_status()
    except:
      pass
    else:
      try:
        newjsd_broadcasts = r1.json()
      except:
        pass
      else:
        if elemExists(newjsd_broadcasts, 'data'):
          # Assumption: broadcasts are listed in descending order (start time)
          for broadcast in newjsd_broadcasts['data']:
            if elemExists(broadcast, 'startdatetime'):
              trackstart = int(datetime.timestamp(datetime.strptime(broadcast['startdatetime'], '%Y-%m-%dT%H:%M:%S')))
              trackend   = int(datetime.timestamp(datetime.strptime(broadcast['stopdatetime'],   '%Y-%m-%dT%H:%M:%S')))
              if trackstart < now + offset and (trackend > now + offset - 1 or trackstart == trackend):
                validbroadcast = True
                break
            if trackstart < now + offset and trackend < now + offset:
              break
          if validbroadcast:
            artUrl = mkStr(broadcast['image_url']).replace('{format}', 'w_1080,h_1080')
            if artUrl:
              print(artUrl)


#
# Copyright (C) 2020 Jean-Francois Dockes
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
    Interface to highresaudio WEB API
"""
import sys
import time
import requests
import json

from upmplgutils import *

# For debugging the details of the HTTP transaction. Note that http.client
# logs to stdout by default, so you need to actually edit the module to log
# to stderr if you want these messages (else they mess up the communication
# with our parent).
# import logging
# import http.client as http_client
# http_client.HTTPConnection.debuglevel = 1
# logging.basicConfig(stream=sys.stderr)
# logging.getLogger().setLevel(logging.DEBUG)
# requests_log = logging.getLogger("requests.packages.urllib3")
# requests_log.setLevel(logging.DEBUG)
# requests_log.propagate = True


# Bogus logging class initially to avoid to change the qobuz module code. Could
# get rid of it here...
class MLog(object):
    def __init__(self):
        self.f = sys.stderr
        self.level = 2

    def _doprint(self, msg):
        print("HRAAPI: %s" % msg, file=self.f)

    def debug(self, msg):
        if self.level >= 3:
            self._doprint(msg)

    def info(self, msg):
        if self.level >= 2:
            self._doprint(msg)

    def warn(self, msg):
        if self.level >= 1:
            self._doprint(msg)


log = MLog()


class HRAAPI(object):

    def __init__(self):
        self.apiUrl = "https://streaming.highresaudio.com:8182/vault3"
        self.session_id = None
        self.user_id = None
        self.user_data = None
        self.logged_on = None
        self.error = None
        self.status_code = None
        self.session = requests.Session()
        self.lang = "en"

    def setlang(self, lang):
        self.lang = lang

    def _api_error_string(self, request, url="", params={}, json=""):
        return (
            "{reason} (code={status_code})\n"
            "url={url}\nparams={params}"
            "\njson={json}".format(
                reason=request.reason,
                status_code=self.status_code,
                url=url,
                params=str(["%s: %s" % (k, v) for k, v in params.items()]),
                json=str(json),
            )
        )

    def _check_ka(self, ka, mandatory, allowed=[]):
        for label in mandatory:
            if not label in ka:
                raise Exception("HRAAPI: missing parameter %s" % label)
        for label in ka:
            if label not in mandatory and label not in allowed:
                raise Exception("HRAAPI: invalid parameter %s" % label)

    def _api_request(self, iparams, uri, **opt):
        """HighresAudio API HTTP get request
        Arguments:
        params:    parameters dictionary
        uri   :    service/method
        opt   :    Optional named parameters: method=GET/POST default GET

        Return None if something went wrong
        Return response data as dictionary on success
        """
        self.error = ""
        self.status_code = None
        url = self.apiUrl + uri
        headers = {}
        params = {}
        dopost = False
        if "method" in opt and opt["method"] == "POST":
            dopost = True
        if self.user_data:
            params = {"userData": self.user_data, "lang": self.lang}
        params.update(iparams)
        if "limit" not in params:
            params["limit"] = 300
        log.info("request: url {}, params: {}".format(url, str(params)))
        r = None
        time1 = time.time()
        try:
            if dopost:
                r = self.session.post(url, data=params, headers=headers)
            else:
                r = self.session.get(url, params=params, headers=headers)
        except:
            self.error = "HTTP request failed"
            log.warn(self.error)
            return None
        time2 = time.time()
        log.info("Request took {:.3f} ms".format((time2 - time1) * 1000.0))

        self.status_code = int(r.status_code)
        if self.status_code != 200:
            self.error = self._api_error_string(r, url, params)
            log.warn(self.error)
            return None
        if not r.content:
            self.error = "Request return no content"
            log.warn(self.error)
            return None
        try:
            response_json = r.json()
        except Exception as e:
            log.warn("Json loads failed to load {}", repr(e))
        status = None
        try:
            status = response_json["response_status"]
        except:
            pass
        if status != "OK":
            self.error = self._api_error_string(r, url, params, response_json)
            log.warn(self.error)
            return None
        log.debug("HRAAPI: response: %s" % response_json)
        return response_json

    def logout(self):
        log.info("logout()")
        self._api_request({}, "/user/logout", method="POST")
        self._clear_login()

    def _renew_session(self):
        log.info("_renew_session()")
        data = self._api_request({}, "/user/keepalive/", method="POST")
        if (
            not data
            or "user_id" not in data
            or "session_id" not in data
            or not data["user_id"]
            or not data["session_id"]
        ):
            self.logged_on = None
            return None
        self.session_id = data["session_id"]
        self.user_id = data["user_id"]
        self.user_data = json.dumps(
            {"user_id": self.user_id, "session_id": self.session_id}
        )
        self.logged_on = time.time()
        return data

    def isloggedin(self):
        log.debug("isloggedin(): user_id %s" % self.user_id)
        # Hra sessions expire after 30mn.
        if self.logged_on:
            now = time.time()
            if now - self.logged_on > 20 * 60:
                self._renew_session()
        return self.logged_on is not None

    def _clear_login(self):
        self.user_session_id = None
        self.user_id = None
        self.logged_on = None

    def user_login(self, **ka):
        self._check_ka(ka, ["username", "password"])
        data = self._api_request(ka, "/user/login", method="POST")
        if (
            not data
            or "user_id" not in data
            or "session_id" not in data
            or not data["user_id"]
            or not data["session_id"]
        ):
            log.warn("login failed")
            self._clear_login()
            return None
        self.session_id = data["session_id"]
        self.user_id = data["user_id"]
        self.user_data = json.dumps(
            {"user_id": self.user_id, "session_id": self.session_id}
        )
        self.logged_on = time.time()
        return data

    def getAllCategories(self):
        if not self.isloggedin():
            return None
        data = self._api_request({}, "/vault/categories/ListAllCategories/")
        if not data or "data" not in data or "results" not in data["data"]:
            return {}
        return [item[1] for item in data["data"]["results"].items()]

    def getAllGenres(self):
        if not self.isloggedin():
            return None
        data = self._api_request({}, "/vault/categories/ListAllGenre")
        if not data or "data" not in data or "results" not in data["data"]:
            return {}
        return data["data"]["results"]

    def listCategoryContent(self, **ka):
        if not self.isloggedin():
            return None
        self._check_ka(ka, ["category"])
        data = self._api_request(ka, "/vault/categories/ListCategorieContent/")
        if not data or "data" not in data or "results" not in data["data"]:
            return {}
        return data["data"]["results"]

    def getAlbumDetails(self, **ka):
        if not self.isloggedin():
            return None
        self._check_ka(ka, ["album_id"])
        data = self._api_request(ka, "/vault/album/")
        if (
            not data
            or "data" not in data
            or "results" not in data["data"]
            or "tracks" not in data["data"]["results"]
        ):
            return None
        return data["data"]["results"]

    def getTrackById(self, **ka):
        if not self.isloggedin():
            return None
        self._check_ka(ka, ["track_id"])
        params = ka
        data = self._api_request(params, "/vault/track/")
        if (
            not data
            or "data" not in data
            or "results" not in data["data"]
            or "tracks" not in data["data"]["results"]
        ):
            return None
        return data["data"]["results"]["tracks"]

    def quickSearch(self, **ka):
        self._check_ka(ka, ["search"])
        log.info("quickSearch: %s" % ka)
        data = self._api_request(ka, "/vault/search/quickSearch/")
        if not data or "data" not in data or not data["data"]:
            return {}
        # The data is a dict key:albumdata, the key can be used as album_id
        # with /vault/album
        return data["data"]

    def searchCategory(self, **ka):
        self._check_ka(ka, ["search", "category"])
        log.info("searchCategory: %s" % ka)
        data = self._api_request(ka, "/vault/SearchInCategory/")
        if not data or "data" not in data or "results" not in data["data"]:
            return {}
        # The data is a dict key:albumdata, the key can be used as album_id
        # with /vault/album
        return data["data"]["results"]

    def getAvailableMoods(self, **ka):
        params = ka
        data = self._api_request(params, "/vault/getEditorPlaylistsMoods/")
        if not data or "data" not in data or "results" not in data["data"]:
            return {}
        return data["data"]["results"]

    def getAvailableGenres(self, **ka):
        params = ka
        data = self._api_request(params, "/vault/getEditorPlaylistsGenres/")
        if not data or "data" not in data or "results" not in data["data"]:
            return {}
        return data["data"]["results"]

    def getAvailableThemes(self, **ka):
        data = self._api_request(ka, "/vault/getEditorPlaylistsThemes/")
        if not data or "data" not in data or "results" not in data["data"]:
            return {}
        return data["data"]["results"]

    def getAllEditorPlaylists(self, **ka):
        data = self._api_request(ka, "/vault/editorPlaylists/")
        if not data or "data" not in data or "results" not in data["data"]:
            return None
        ret = data["data"]["results"]
        return ret

    def getEditorPlaylist(self, **ka):
        if not self.isloggedin():
            return None
        self._check_ka(ka, ["id"])
        data = self._api_request(ka, "/vault/getSingleEditorPlaylists/")
        if (
            not data
            or "data" not in data
            or "results" not in data["data"]
            or "tracks" not in data["data"]["results"][0]
        ):
            log.info("getEditorPlaylist: RETURNING NONE for id %s" % ka["id"])
            return None
        return data["data"]["results"][0]

    def listAllUserPlaylists(self, **ka):
        if not self.isloggedin():
            return {}
        data = self._api_request(ka, "/user/ListAllUserPlaylists")
        if not data or "data" not in data or "data" not in data["data"]:
            return None
        ret = data["data"]["data"]
        return ret

    def getUserPlaylist(self, **ka):
        if not self.isloggedin():
            return None
        self._check_ka(ka, ["playlist_id"])
        data = self._api_request(ka, "/user/ListSingleUserPlaylist/")
        if not data or "data" not in data or "data" not in data["data"]:
            return None
        return data["data"]["data"]

    def listAllUserAlbums(self, **ka):
        if not self.isloggedin():
            return {}
        params = ka
        params["limit"] = 1000
        data = self._api_request(params, "/user/list/MyAlbum")
        if not data or "data" not in data or "results" not in data["data"]:
            return None
        return data["data"]["results"]

    def listAllUserTracks(self, **ka):
        if not self.isloggedin():
            return {}
        params = ka
        params["limit"] = 2000
        data = self._api_request(params, "/user/list/MyTracks")
        if (
            not data
            or "data" not in data
            or "data" not in data["data"]
            or not "results" in data["data"]["data"]
        ):
            return None
        ret = data["data"]["data"]["results"]
        return ret



#
# Copyright (C) 2017 J.F.Dockes
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Manage the 'untagged' section of the tree.  Our selection is made of
# all tracks without a 'title' field. This is different from what
# Minimserver does (I think that more or less any field present makes
# a track not untagged for minim).
#
# Initialization filters the untagged tracks and creates a vector of
# indexes into the global doc vector.
#
# Object Id prefix: 0$uprcl$untagged
#
# Obect id inside the section: $u<idx> where <idx> is the document index
#  inside the global document vector.

import os
import shlex
import sys

from upmplgutils import uplog, direntry
from uprclutils import rcldoctoentry
import uprclinit


class Untagged(object):
    def __init__(self, rcldocs, httphp, pathprefix):
        self._idprefix = "0$uprcl$untagged"
        self._httphp = httphp
        self._pprefix = pathprefix
        self.recoll2untagged(rcldocs)

    # Create the untagged entries static vector by filtering the global
    # doc vector, storing the indexes of all tracks without a title
    # field. We keep a reference to the doc vector.
    def recoll2untagged(self, rcldocs):
        # The -1 entry is because we use index 0 for our root.
        self.utidx = [-1]

        for docidx in range(len(rcldocs)):
            doc = rcldocs[docidx]
            if doc["mtype"] == "inode/directory" or doc["mtype"] == "audio/x-mpegurl":
                continue
            if not doc["title"]:
                self.utidx.append(docidx)

    # Compute index into our entries vector by 'parsing' the objid.
    def _objidtoidx(self, pid):
        if not pid.startswith(self._idprefix):
            raise Exception("untagged:browse: bad pid %s" % pid)

        idx = pid[len(self._idprefix) :]
        if not idx:
            # Browsing the root.
            idx = 0
        else:
            if idx[1] != "u":
                raise Exception("untagged:browse: called on bad objid %s" % pid)
            idx = int(idx[2:])

        if idx >= len(self.utidx):
            raise Exception("untagged:browse: bad pid %s" % pid)

        return idx

    # Return entry to be created in the top-level directory ([untagged]).
    def rootentries(self, pid):
        return [
            direntry(pid + "untagged", pid, "[untagged]"),
        ]

    # Browse method
    # objid is like 0$uprcl$untagged$u<index>
    # flag is meta or children.
    def browse(self, pid, flag, offset, count):
        idx = self._objidtoidx(pid)

        entries = []
        rcldocs = uprclinit.getTree("folders").rcldocs()
        if idx == 0:
            # Browsing root
            if flag == "meta":
                # Root itself
                return self.rootentries("0$uprcl$")
            else:
                # Root children
                for i in range(len(self.utidx))[1:]:
                    doc = rcldocs[self.utidx[i]]
                    # uplog(f"UNTAGGED: {i} -> {doc['url']}")
                    id = self._idprefix + "$u" + str(i)
                    e = rcldoctoentry(id, pid, self._httphp, self._pprefix, doc)
                    if e:
                        entries.append(e)
        else:
            # Non root: only items in there. flag needs to be 'meta'
            doc = rcldocs[self.utidx[idx]]
            id = self._idprefix + "$u" + str(idx)
            e = rcldoctoentry(id, pid, self._httphp, self._pprefix, doc)
            if e:
                entries.append(e)

        return sorted(entries, key=lambda entry: entry["tt"].lower())



# Copyright (C) 2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from metadata import Metadata
from metadata_model import SongMetadataModel
from typing import Optional
from typing import Any
import datetime


class SongMetadata(Metadata):

    def get_value(self, song_metadata_model: SongMetadataModel) -> Optional[Any]:
        return self._get(song_metadata_model.column_name)

    def set_value(self, song_metadata_model: SongMetadataModel, value: Any):
        return self._set(song_metadata_model.column_name, value)

    @property
    def song_id(self) -> str:
        return self.get_value(SongMetadataModel.SONG_ID)

    @property
    def album_id(self) -> str:
        return self.get_value(SongMetadataModel.SONG_ALBUM_ID)

    @property
    def created_timestamp(self) -> datetime.datetime:
        return self.get_value(SongMetadataModel.CREATED_TIMESTAMP)

    @property
    def updated_timestamp(self) -> datetime.datetime:
        return self.get_value(SongMetadataModel.UPDATED_TIMESTAMP)



##########################################################################
# Copyright 2009 Carlos Ribeiro
#
# This file is part of Radio Tray
#
# Radio Tray is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 1 of the License, or
# (at your option) any later version.
#
# Radio Tray is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radio Tray.  If not, see <http://www.gnu.org/licenses/>.
#
##########################################################################
import sys
from urllib.request import Request as UrlRequest
from urllib.request import urlopen as urlUrlopen
import logging
from common import USER_AGENT

class PlsPlaylistDecoder:
    def __init__(self):
        self.log = logging.getLogger('upmpdcli')
        self.log.debug('PLS playlist decoder')
        
    def isStreamValid(self, contentType, firstBytes):
        if 'audio/x-scpls' in contentType or \
           'application/pls+xml' in contentType or \
           firstBytes.strip().lower().startswith(b'[playlist]'):
            self.log.info('Stream is readable by PLS Playlist Decoder')
            return True
        else:
            return False

    def extractPlaylist(self,  url):
            
            self.log.info('Downloading playlist...')
            
            req = UrlRequest(url)
            req.add_header('User-Agent', USER_AGENT)
            f = urlUrlopen(req)
            str = f.read()
            f.close()
            
            self.log.info('Playlist downloaded')
            self.log.info('Decoding playlist...')
            
            playlist = []
            lines = str.splitlines()
            for line in lines:
                if line.startswith(b"File") == True:
                        list = line.split(b"=", 1)
                        playlist.append(list[1])
     
            
            return playlist
            
            



# Copyright (C) 2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# source: https://musicbrainz.org/doc/Release_Group/Type

from enum import Enum
from typing import Callable


class _ReleaseTypeDefinition:

    def __init__(self, value_list: list[str], reference_value: str = None, display_value: str = None):
        self.__value_list: list[str] = value_list
        self.__reference_value: str = reference_value if reference_value else value_list[0]
        self.__display_value: str = display_value if display_value else self.__reference_value.title()

    @property
    def value_list(self) -> list[str]:
        return self.__value_list

    @property
    def reference_value(self) -> str:
        return self.__reference_value

    @property
    def display_value(self) -> str:
        return self.__display_value


def _words_to_array(word_list: list[str]) -> list[str]:
    lst: list[str] = []
    lst.append("/".join(word_list))
    lst.append("-".join(word_list))
    lst.append(" ".join(word_list))
    return lst


class PrimaryReleaseType(Enum):
    ALBUM = _ReleaseTypeDefinition(["album"])
    SINGLE = _ReleaseTypeDefinition(["single"])
    EP = _ReleaseTypeDefinition(["ep"])
    BROADCAST = _ReleaseTypeDefinition(["broadcast"])
    OTHER = _ReleaseTypeDefinition(["other"])

    @property
    def value_list(self) -> list[str]:
        return self.value.value_list

    @property
    def reference_value(self) -> str:
        return self.value.reference_value

    @property
    def display_value(self) -> str:
        return self.value.display_value


class SecondaryReleaseType(Enum):
    COMPILATION = _ReleaseTypeDefinition(["compilation"])
    SOUNDTRACK = _ReleaseTypeDefinition(["soundtrack"])
    SPOKEN_WORD = _ReleaseTypeDefinition(["spokenword"])
    INTERVIEW = _ReleaseTypeDefinition(["interview"])
    AUDIO_BOOK = _ReleaseTypeDefinition(["audiobook"], display_value="Audiobook")
    AUDIO_DRAMA = _ReleaseTypeDefinition(["audiodrama"], display_value="Audio drama")
    LIVE = _ReleaseTypeDefinition(["live"])
    REMIX = _ReleaseTypeDefinition(["remix"])
    DJ_MIX = _ReleaseTypeDefinition(_words_to_array(["dj", "mix"]), display_value="DJ-mix")
    MIXTAPE_STREET = _ReleaseTypeDefinition(_words_to_array(["mixtape", "street"]), display_value="Mixtape/Street")
    DEMO = _ReleaseTypeDefinition(["demo"])
    FIELD_RECORDING = _ReleaseTypeDefinition(_words_to_array(["field", "recording"]), display_value="Field recording")

    @property
    def value_list(self) -> list[str]:
        return self.value.value_list

    @property
    def reference_value(self) -> str:
        return self.value.reference_value

    @property
    def display_value(self) -> str:
        return self.value.display_value


def match_primary_release_type(t: str) -> PrimaryReleaseType:
    if not t:
        return False
    low_t: str = t.lower()
    prt: PrimaryReleaseType
    for prt in PrimaryReleaseType:
        if low_t in prt.value_list:
            return prt
    return None


def match_secondary_release_type(t: str) -> SecondaryReleaseType:
    if not t:
        return False
    low_t: str = t.lower()
    srt: SecondaryReleaseType
    for srt in SecondaryReleaseType:
        if low_t in srt.value_list:
            return srt
    return None


def is_primary_release_type(t: str) -> bool:
    return match_primary_release_type(t) is not None


def is_secondary_release_type(t: str) -> bool:
    return match_secondary_release_type(t) is not None


def any_primary_release_type(value_list: list[str]) -> bool:
    return len(extract_release_types(
        value_list=value_list,
        type_matcher=is_primary_release_type)) > 0


def any_secondary_release_type(value_list: list[str]) -> bool:
    return len(extract_release_types(
        value_list=value_list,
        type_matcher=is_secondary_release_type)) > 0


def display_value(value: str) -> str:
    if not value:
        return None
    primary: PrimaryReleaseType = match_primary_release_type(value)
    if primary:
        return primary.display_value
    secondary: SecondaryReleaseType = match_secondary_release_type(value)
    if secondary:
        return secondary.display_value
    return value.title()


def extract_release_types(value_list: list[str], type_matcher: Callable[[str], bool]) -> list[str]:
    res: list[str] = []
    v: str
    for v in (value_list if value_list else []):
        if type_matcher(v):
            res.append(v)
    return res


def sanitize_release_types(
        value_list: list[str],
        fallback_primary: PrimaryReleaseType = PrimaryReleaseType.ALBUM,
        print_function: Callable[[str], None] = None,
        album_id: str = None,
        album_title: str = None,
        album_artist: str = None) -> list[str]:
    # correct some common errors.
    primaries: list[PrimaryReleaseType] = []
    secondaries: list[SecondaryReleaseType] = []
    others: list[str] = []
    res: list[str] = []
    v: str
    for v in value_list if value_list else []:
        # skip empty values
        if not v:
            continue
        # primary?
        p: PrimaryReleaseType = match_primary_release_type(v)
        if p and p not in primaries:
            primaries.append(p)
        # secondary?
        s: SecondaryReleaseType = not p and match_secondary_release_type(v)
        if s and s not in secondaries:
            secondaries.append(s)
        if not (s or p):
            # other!
            if v not in others:
                others.append(v)
    # only a secondary value, and maybe others?
    if len(primaries) == 0 and len(secondaries) > 0:
        if fallback_primary:
            res.append(fallback_primary.reference_value)
        res.append(secondaries[0].reference_value)
        res.extend(others)
        if print_function:
            print_function(f"sanitize_release_types for album_id [{album_id}] "
                           f"title [{album_title}] "
                           f"by [{album_artist}] "
                           f"values [{value_list}] -> "
                           f"[{res}] (missing primary)")
    elif len(primaries) == 0 and len(secondaries) == 0 and len(others) == 0:
        if fallback_primary:
            res.append(fallback_primary.reference_value)
    else:
        # push everything.
        res.extend(map(lambda p: p.reference_value, primaries))
        res.extend(map(lambda p: p.reference_value, secondaries))
        res.extend(others)
    return res


# tests
album_is_primary: bool = any_primary_release_type(["Album", "ep"])
if not album_is_primary:
    raise Exception("Album should match as primary")
compilation_is_secondary: bool = any_secondary_release_type(["Album", "Compilation"])
if not compilation_is_secondary:
    raise Exception("Compilation should match as secondary")
# extract correctly
album_extract: list[str] = extract_release_types(
    value_list=["Demo", "Album"],
    type_matcher=is_primary_release_type)[0]
if not PrimaryReleaseType.ALBUM == match_primary_release_type(album_extract):
    raise Exception("Album should be extract as album regardless of the position")
# sanitize cases
sanitized_1: list[str] = sanitize_release_types(value_list=["compilation"])
if not ["album", "compilation"] == sanitized_1:
    raise Exception("Sanitization case #1 failed")
sanitized_2: list[str] = sanitize_release_types(value_list=["album", "live"])
if not ["album", "live"] == sanitized_2:
    raise Exception("Sanitization case #2 failed")
sanitized_3: list[str] = sanitize_release_types(value_list=["compilation", "some_other"])
if not ["album", "compilation", "some_other"] == sanitized_3:
    raise Exception("Sanitization case #3 failed")
wrong_4: list[str] = ["DeMo", "album"]
sanitized_4: list[str] = sanitize_release_types(value_list=wrong_4)
if not ["album", "demo"] == sanitized_4:
    raise Exception("Sanitization case #4 failed")
# secondary with separator
if not SecondaryReleaseType.MIXTAPE_STREET == match_secondary_release_type("mixtape-sTrEeT"):
    raise Exception(f"Match for {SecondaryReleaseType.MIXTAPE_STREET.reference_value} failed")



# Copyright (C) 2024 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from context_key import ContextKey


class Context:

    def __init__(self):
        self._dict: dict[str, any] = dict()

    def __not_null(self, key: ContextKey):
        if not key:
            raise Exception("Provided key is None")

    def __raise_missing_key(self, key_str: str):
        raise Exception(f"Missing key [{key_str}]")

    def __not_in(self, key: ContextKey):
        self.__not_null(key)
        if key in self._dict:
            raise Exception(f"Duplicate key [{key.name}]")

    def contains(self, key: ContextKey) -> bool:
        self.__not_null(key)
        return key.name in self._dict

    def add(self, key: ContextKey, value: any, allow_update: bool = False):
        if allow_update:
            self.__not_null(key)
        else:
            self.__not_in(key)
        self._dict[key.name] = value

    def update(self, key: ContextKey, value: any):
        self.add(key=key, value=value, allow_update=True)

    def increment(self, key: ContextKey):
        v: any = self.get(key=key)
        if not isinstance(v, int):
            raise Exception(f"Item at key [{key.name}] is not a int")
        self.update(key=key, value=int(v) + 1)

    def get(self, key: ContextKey, allow_empty: bool = True):
        self.__not_null(key)
        if not allow_empty and key.name not in self._dict:
            self.__raise_missing_key(key.name)
        return self._dict[key.name] if key.name in self._dict else key.default_value

    def remove(self, key: ContextKey, allow_empty: bool = True):
        if key.name in self.__dict:
            del self._dict[key.name]
        else:
            if not allow_empty:
                raise Exception(f"Missing key [{key.name}]")

    def __get_dict(self, dict_key: ContextKey, allow_empty: bool = True) -> dict[str, any]:
        select_dict: dict[str, any] = self.get(key=dict_key)
        if not select_dict:
            if not allow_empty:
                raise Exception(f"Key [{dict_key.name}] not found")
            select_dict = dict()
            self.add(key=dict_key, value=select_dict)
        else:
            # is it a dictionary?
            if not isinstance(select_dict, dict):
                raise Exception(f"Element by key [{dict_key.name}] is not a dictionary")
        return select_dict

    def dict_add(
            self,
            dict_key: ContextKey,
            entry_key: str,
            entry_value: any,
            allow_empty: bool = True,
            allow_update: bool = False):
        # check if key exists
        select_dict: dict[str, any] = self.__get_dict(dict_key=dict_key, allow_empty=allow_empty)
        if entry_key in select_dict and not allow_update:
            raise Exception(f"Duplicate key [{entry_key}]")
        select_dict[entry_key] = entry_value

    def dict_update(
            self,
            dict_key: ContextKey,
            entry_key: str,
            entry_value: any,
            allow_empty: bool = True):
        self.dict_add(
            dict_key=dict_key,
            entry_key=entry_key,
            entry_value=entry_value,
            allow_empty=allow_empty,
            allow_update=True)

    def dict_remove(
            self,
            dict_key: ContextKey,
            entry_key: str,
            allow_empty: bool = True,
            allow_missing: bool = True):
        select_dict: dict[str, any] = self.__get_dict(dict_key=dict_key, allow_empty=allow_empty)
        if entry_key in select_dict:
            del select_dict[entry_key]
        else:
            if not allow_missing:
                raise Exception(f"Key [{entry_key}] not found")

    def dict_get(
            self,
            dict_key: ContextKey,
            entry_key: str,
            allow_empty: bool = True,
            allow_missing: bool = True,
            default_value: any = None) -> bool:
        select_dict: dict[str, any] = self.__get_dict(dict_key=dict_key, allow_empty=allow_empty)
        if entry_key not in select_dict:
            if not allow_missing:
                raise Exception(f"Key [{entry_key}] not found")
            return default_value
        else:
            return select_dict[entry_key]

    def dict_contains(
            self,
            dict_key: ContextKey,
            entry_key: str,
            allow_empty: bool = True) -> bool:
        select_dict: dict[str, any] = self.__get_dict(dict_key=dict_key, allow_empty=allow_empty)
        return entry_key in select_dict



# Copyright (C) 2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from tidalapi.album import Album as TidalAlbum
from tidalapi.artist import Artist as TidalArtist

from datetime import datetime
from persistence import AlbumMetadata
import tidal_util
import persistence
import copy
import typing
import cmdtalkplugin

# Func name to method mapper
dispatcher = cmdtalkplugin.Dispatch()
# Pipe message handler
msgproc = cmdtalkplugin.Processor(dispatcher)


def __albumToAlbumMetadata(album: TidalAlbum) -> AlbumMetadata:
    result: AlbumMetadata = AlbumMetadata()
    result.album_id = album.id
    result.album_name = album.name
    result.artist_id = album.artist.id if album.artist else None
    result.artist_name = album.artist.name if album.artist else None
    result.explicit = 1 if album.explicit else 0
    result.release_date = album.release_date
    result.available_release_date = album.available_release_date
    result.image_url = tidal_util.get_image_url(album)
    album_audio_modes: list[str] = album.audio_modes
    result.audio_modes = ",".join(album_audio_modes) if album_audio_modes else None
    result.audio_quality = album.audio_quality
    album_media_metadata_tags: list[str] = album.media_metadata_tags
    result.media_metadata_tags = ",".join(album_media_metadata_tags) if album_media_metadata_tags else None
    result.album_duration = album.duration
    result.num_volumes = album.num_volumes
    result.num_tracks = album.num_tracks
    return result


class AlbumAdapter:

    id: str = None
    name: str = None
    artist_id: str = None
    artist_name: str = None
    artist_list: list[TidalArtist] = None
    explicit: bool = False
    release_date: datetime = None
    available_release_date: datetime = None
    image_url: str = None
    audio_modes: list[str] = None
    audio_quality: str = None
    media_metadata_tags: list[str] = None
    version: str = None
    universal_product_number: int = None
    copyright: str = None
    duration: int = None
    num_tracks: int = None
    num_volumes: int = None

    @property
    def year(self) -> int:
        return self.available_release_date.year if self.available_release_date else None


def tidal_album_to_adapter(tidal_album: TidalAlbum) -> AlbumAdapter:
    persistence.store_album_metadata(__albumToAlbumMetadata(tidal_album))
    album_adapter: AlbumAdapter = AlbumAdapter()
    album_adapter.id = tidal_album.id
    album_adapter.name = tidal_album.name
    album_adapter.artist_id = tidal_album.artist.id if tidal_album.artist else None
    album_adapter.artist_name = tidal_album.artist.name if tidal_album.artist else None
    album_adapter.artist_list = []
    curr_artist: TidalArtist
    for curr_artist in tidal_album.artists if tidal_album.artists else []:
        album_adapter.artist_list.append(curr_artist)
    album_adapter.explicit = tidal_album.explicit
    album_adapter.release_date = tidal_album.release_date
    album_adapter.available_release_date = tidal_album.available_release_date
    album_adapter.image_url = tidal_util.get_image_url(tidal_album)
    album_adapter.audio_modes = copy.deepcopy(tidal_album.audio_modes)
    album_adapter.audio_quality = tidal_album.audio_quality
    album_adapter.media_metadata_tags = copy.deepcopy(tidal_album.media_metadata_tags)
    album_adapter.version = tidal_album.version
    album_adapter.universal_product_number = tidal_album.universal_product_number
    album_adapter.copyright = tidal_album.copyright
    album_adapter.duration = tidal_album.duration
    album_adapter.num_volumes = tidal_album.num_volumes
    album_adapter.num_tracks = tidal_album.num_tracks
    return album_adapter


def __album_metadata_to_adapter(album_metadata: AlbumMetadata) -> AlbumAdapter:
    album_adapter: AlbumAdapter = AlbumAdapter()
    album_adapter.id = album_metadata.album_id
    album_adapter.name = album_metadata.album_name
    album_adapter.artist_id = album_metadata.artist_id
    album_adapter.artist_name = album_metadata.artist_name
    album_adapter.explicit = True if album_metadata.explicit == 1 else False
    album_adapter.release_date = album_metadata.release_date
    album_adapter.available_release_date = album_metadata.available_release_date
    album_adapter.image_url = album_metadata.image_url
    album_adapter.audio_modes = (album_metadata.audio_modes.split(",")
                                 if album_metadata.audio_modes
                                 else None)
    album_adapter.audio_quality = album_metadata.audio_quality
    album_adapter.media_metadata_tags = (album_metadata.media_metadata_tags.split(",")
                                         if album_metadata.media_metadata_tags
                                         else None)
    album_adapter.duration = album_metadata.album_duration
    album_adapter.num_volumes = album_metadata.num_volumes
    album_adapter.num_tracks = album_metadata.num_tracks
    return album_adapter


def album_adapter_by_album_id(
        album_id: str,
        tidal_album_loader: typing.Callable[[str], TidalAlbum]) -> AlbumAdapter:
    album_metadata: AlbumMetadata = persistence.get_album_metadata(album_id=album_id)
    # msgproc.log(f"album_adapter_by_album_id [{album_id}] cache hit: [{'yes' if album_metadata else 'no'}]")
    if album_metadata:
        return __album_metadata_to_adapter(album_metadata)
    # load the album if no metadata is available
    tidal_album: TidalAlbum = tidal_album_loader(album_id)
    # convert to adapter (performs caching)
    return tidal_album_to_adapter(tidal_album) if tidal_album else None



# Copyright (C) 2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from metadata import Metadata
from metadata_model import ArtistMetadataModel
from typing import Optional
from typing import Any
import datetime


class ArtistMetadata(Metadata):

    def get_value(self, artist_metadata_model: ArtistMetadataModel) -> Optional[Any]:
        return self._get(artist_metadata_model.column_name)

    def set_value(self, artist_metadata_model: ArtistMetadataModel, value: Any):
        return self._set(artist_metadata_model.column_name, value)

    @property
    def artist_id(self) -> str:
        return self.get_value(ArtistMetadataModel.ARTIST_ID)

    @property
    def artist_name(self) -> str:
        return self.get_value(ArtistMetadataModel.ARTIST_NAME)

    @property
    def artist_musicbrainz_id(self) -> str:
        return self.get_value(ArtistMetadataModel.ARTIST_MB_ID)

    @property
    def artist_album_count(self) -> int:
        return self.get_value(ArtistMetadataModel.ARTIST_ALBUM_COUNT)

    @property
    def artist_cover_art(self) -> str:
        return self.get_value(ArtistMetadataModel.ARTIST_COVER_ART)

    @property
    def artist_media_type(self) -> str:
        return self.get_value(ArtistMetadataModel.ARTIST_MEDIA_TYPE)

    @property
    def artist_sort_name(self) -> str:
        return self.get_value(ArtistMetadataModel.ARTIST_SORT_NAME)

    @property
    def created_timestamp(self) -> datetime.datetime:
        return self.get_value(ArtistMetadataModel.CREATED_TIMESTAMP)

    @property
    def updated_timestamp(self) -> datetime.datetime:
        return self.get_value(ArtistMetadataModel.UPDATED_TIMESTAMP)

    def update(
            self,
            artist_name: str,
            artist_musicbrainz_id: str,
            artist_album_count: int = None,
            artist_cover_art: str = None,
            artist_media_type: str = None,
            artist_sort_name: str = None):
        any_update: bool = False
        if artist_name and len(artist_name) > 0:
            self.set_value(ArtistMetadataModel.ARTIST_NAME, artist_name)
            any_update = True
        if artist_musicbrainz_id and len(artist_musicbrainz_id) > 0:
            self.set_value(ArtistMetadataModel.ARTIST_MB_ID, artist_musicbrainz_id)
            any_update = True
        if artist_album_count:
            self.set_value(ArtistMetadataModel.ARTIST_ALBUM_COUNT, artist_album_count)
            any_update = True
        if artist_cover_art:
            self.set_value(ArtistMetadataModel.ARTIST_COVER_ART, artist_cover_art)
            any_update = True
        if artist_media_type:
            self.set_value(ArtistMetadataModel.ARTIST_MEDIA_TYPE, artist_media_type)
            any_update = True
        if artist_sort_name:
            self.set_value(ArtistMetadataModel.ARTIST_SORT_NAME, artist_sort_name)
            any_update = True
        if any_update:
            self.set_value(ArtistMetadataModel.UPDATED_TIMESTAMP, datetime.datetime.now())



# Copyright (C) 2023,2024,2025,2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from item_identifier import ItemIdentifier
from html import escape

import codec

import json

def __escape_objid(value: str) -> str:
    return escape(value, quote = True)

def create_objid(objid, id: str) -> str:
    return objid + "/" + __escape_objid(id)

def create_id_from_identifier(identifier: ItemIdentifier) -> str:
    return codec.encode(json.dumps(identifier.getDictionary()))




# Copyright (C) 2023,2024,2025,2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from subsonic_connector.response import Response
from subsonic_connector.album_list import AlbumList
from subsonic_connector.list_type import ListType
from subsonic_connector.playlists import Playlists
from subsonic_connector.playlist import Playlist
from subsonic_connector.starred import Starred
from subsonic_connector.artist import Artist
from subsonic_connector.album import Album
from subsonic_connector.song import Song

from persistence_tuple import ArtistAlbumCoverArt
from tag_type import TagType
from retrieved_art import RetrievedArt

from tag_to_entry_context import TagToEntryContext

import connector_provider
import request_cache
import subsonic_util
import cache_actions
import persistence
import config
import constants

import secrets

from typing import Callable
from msgproc_provider import msgproc


def _get_cover_art_from_res_album_list(response: Response[AlbumList], random: bool = False) -> RetrievedArt:
    if not response.isOk() or len(response.getObj().getAlbums()) == 0:
        return None
    album_list: list[Album] = response.getObj().getAlbums()
    return __get_cover_art_from_album_list(album_list=album_list, random=random)


def __get_cover_art_from_album_list(album_list: list[Album], random: bool = False) -> RetrievedArt:
    album_list = _filter_album_without_cover_art(album_list=album_list)
    if not album_list or len(album_list) == 0:
        return None
    album: Album = (secrets.choice(album_list)
                    if random
                    else album_list[0])
    album_url: str = subsonic_util.build_cover_art_url(item_id=album.getCoverArt()) if album else None
    return RetrievedArt(art_url=album_url) if album_url else None


def _filter_album_without_cover_art(album_list: list[Album]) -> list[Album]:
    return (list(filter(lambda x: x.getCoverArt() is not None and len(x.getCoverArt()) > 0, album_list))
            if album_list
            else [])


def consume_random_album_for_cover_art(tag_to_entry_context: TagToEntryContext = None) -> Album:
    if tag_to_entry_context is None:
        msgproc.log("consume_random_album_for_cover_art cannot return an album because a context is not available")
        return None
    random_album_list: list[Album] = tag_to_entry_context.random_album_list
    msgproc.log(f"Albums available in tagToEntryContext [{len(random_album_list)}] ...")
    select_album: Album = None
    while not select_album and len(random_album_list) > 0:
        select: Album = random_album_list.pop(0)
        # does the album have a cover art?
        if select.getCoverArt() is not None:
            msgproc.log(f"consume_random_album_for_cover_art consumed album [{select.getId()}] [{select.getTitle()}] "
                        f"by [{subsonic_util.get_album_display_artist(album=select)}]")
            return select
        else:
            msgproc.log(f"Skipping album [{select.getId()}] [{select.getTitle()}] "
                        f"from [{subsonic_util.get_album_display_artist(album=select)}] (no cover art)")
    msgproc.log("consume_random_album_for_cover_art cannot not return an album")
    return None


def group_albums_art_retriever(tag_to_entry_context: TagToEntryContext = None) -> RetrievedArt:
    msgproc.log(f"group_albums_art_retriever tag_to_entry_context is set: [{tag_to_entry_context is not None}]")
    # try in favorites (pick random), if allowed
    art: RetrievedArt = (favourite_albums_art_retriever()
                         if config.get_config_param_as_bool(constants.ConfigParam.ALLOW_FAVORITES_FOR_FRONT_PAGE_TAGS)
                         else None)
    # else random album from context, if available
    random_album: Album = consume_random_album_for_cover_art(tag_to_entry_context=tag_to_entry_context)
    # set RetrievedArt if album is found
    art = (RetrievedArt(art_url=subsonic_util.build_cover_art_url(item_id=random_album.getCoverArt()))
           if random_album
           else None)
    # else random
    if not art:
        art = random_albums_art_retriever()
    return art


def group_artists_art_retriever(tag_to_entry_context: TagToEntryContext = None) -> RetrievedArt:
    # try in favorites (pick random)
    art: RetrievedArt = (favourite_artist_art_retriever()
                         if config.get_config_param_as_bool(constants.ConfigParam.ALLOW_FAVORITES_FOR_FRONT_PAGE_TAGS)
                         else None)
    # else random album from context, if available
    random_album: Album = consume_random_album_for_cover_art(tag_to_entry_context=tag_to_entry_context)
    # set RetrievedArt if album is found
    art = (RetrievedArt(art_url=subsonic_util.build_cover_art_url(item_id=random_album.getCoverArt()))
           if random_album
           else None)
    if not art:
        art = random_albums_art_retriever()
    return art


def group_songs_art_retriever(tag_to_entry_context: TagToEntryContext = None) -> RetrievedArt:
    # try in favorites (pick random)
    art: RetrievedArt = (_art_for_favourite_song(random=True)
                         if config.get_config_param_as_bool(constants.ConfigParam.ALLOW_FAVORITES_FOR_FRONT_PAGE_TAGS)
                         else None)
    # else random album from context, if available
    random_album: Album = consume_random_album_for_cover_art(tag_to_entry_context=tag_to_entry_context)
    # set RetrievedArt if album is found
    art = (RetrievedArt(art_url=subsonic_util.build_cover_art_url(item_id=random_album.getCoverArt()))
           if random_album
           else None)
    # else random
    if not art:
        art = random_albums_art_retriever()
    return art


def newest_albums_art_retriever(tag_to_entry_context: TagToEntryContext = None) -> RetrievedArt:
    response: Response[AlbumList] = request_cache.get_first_newest_album_list()
    return _get_cover_art_from_res_album_list(response=response, random=False)


def recently_added_albums_art_retriever(tag_to_entry_context: TagToEntryContext = None) -> RetrievedArt:
    response: Response[AlbumList] = connector_provider.get().getNewestAlbumList(size=1)
    return _get_cover_art_from_res_album_list(response=response)


def random_albums_art_retriever(tag_to_entry_context: TagToEntryContext = None) -> RetrievedArt:
    random_album_from_context: Album = consume_random_album_for_cover_art(tag_to_entry_context=tag_to_entry_context)
    if random_album_from_context:
        return RetrievedArt(art_url=subsonic_util.build_cover_art_url(item_id=random_album_from_context.getCoverArt()))
    response: Response[AlbumList] = request_cache.get_random_album_list(size=config.get_items_per_page())
    return _get_cover_art_from_res_album_list(response=response, random=True)


def recently_played_albums_art_retriever(tag_to_entry_context: TagToEntryContext = None) -> RetrievedArt:
    response: Response[AlbumList] = connector_provider.get().getAlbumList(ltype=ListType.RECENT, size=1)
    return _get_cover_art_from_res_album_list(response=response)


def highest_rated_albums_art_retriever(tag_to_entry_context: TagToEntryContext = None) -> RetrievedArt:
    response: Response[AlbumList] = connector_provider.get().getAlbumList(ltype=ListType.HIGHEST, size=1)
    return _get_cover_art_from_res_album_list(response=response)


def most_played_albums_art_retriever(tag_to_entry_context: TagToEntryContext = None) -> RetrievedArt:
    response: Response[AlbumList] = connector_provider.get().getAlbumList(ltype=ListType.FREQUENT, size=1)
    return _get_cover_art_from_res_album_list(response=response)


def get_album_cover_art_uri_by_artist(artist: Artist, force_save: bool = False) -> str:
    # best option, the artist has its coverArt
    artist_cover_art: str = subsonic_util.get_artist_cover_art(artist=artist)
    if artist_cover_art:
        return subsonic_util.build_cover_art_url(item_id=artist_cover_art, force_save=force_save)
    # maybe look at the albums if available
    album_list: list[Album] = artist.getAlbumList()
    if config.get_verbose_logging():
        msgproc.log(f"get_album_cover_art_uri_by_artist [{artist.getId()}] [{artist.getName()}] "
                    f"album_list len is [{len(album_list)}]")
    if not album_list:
        album_list = []
    non_appearances: list[Album] = subsonic_util.get_artist_albums_as_appears_on(
        artist_id=artist.getId(),
        album_list=album_list,
        opposite=True)
    appearances: list[Album] = subsonic_util.get_artist_albums_as_appears_on(
        artist_id=artist.getId(),
        album_list=album_list,
        opposite=False)
    if len(non_appearances) > 0:
        subsonic_util.sort_albums_by_date(album_list=non_appearances)
        # take first with a coverArt
        first_cover_art: str = __get_first_cover_art_from_album_list(album_list=non_appearances)
        if first_cover_art:
            return subsonic_util.build_cover_art_url(item_id=first_cover_art)
    # fallback into appearances
    if len(appearances) > 0:
        subsonic_util.sort_albums_by_date(album_list=appearances)
        # take first with a coverArt
        first_cover_art: str = __get_first_cover_art_from_album_list(album_list=appearances)
        if first_cover_art:
            return subsonic_util.build_cover_art_url(item_id=first_cover_art)
    # otherwise we fallback to other options
    return get_album_cover_art_url_by_artist_id(artist_id=artist.getId())


def get_album_cover_art_by_artist_id(artist_id: str, skip_artist_metadata_cache: bool = False) -> str:
    verbose: bool = config.get_verbose_logging()
    if verbose:
        msgproc.log(f"get_album_cover_art_by_artist_id for [{artist_id}] "
                    f"skip_artist_metadata_cache [{skip_artist_metadata_cache}]")
    artist_metadata: persistence.ArtistMetadata = (persistence.get_artist_metadata(artist_id=artist_id)
                                                   if not skip_artist_metadata_cache
                                                   else None)
    if artist_metadata and artist_metadata.artist_cover_art:
        # found in cache.
        if verbose:
            msgproc.log(f"get_album_cover_art_by_artist_id metadata cache hit for [{artist_id}] -> "
                        f"[{'yes' if artist_metadata else 'no'}]")
        return artist_metadata.artist_cover_art
    # second option is album cover art by artist id table
    album_cover_art_list: list[ArtistAlbumCoverArt] = persistence.get_cover_art_list_by_artist_id(artist_id=artist_id)
    # filter by cover art set
    album_cover_art_list_with_cover: list[ArtistAlbumCoverArt] = list(filter(
        lambda x: x.cover_art,
        album_cover_art_list))
    if len(album_cover_art_list_with_cover) > 0:
        if verbose:
            msgproc.log(f"get_album_cover_art_by_artist_id for [{artist_id}] "
                        f"skip_artist_metadata_cache [{skip_artist_metadata_cache}] "
                        f"found [{len(album_cover_art_list)}] "
                        "using get_cover_art_list_by_artist_id")
        return secrets.choice(album_cover_art_list_with_cover).cover_art
    album_cover_art_list_without_cover: list[ArtistAlbumCoverArt] = list(filter(
            lambda x: x.cover_art is None,
            album_cover_art_list))
    if len(album_cover_art_list_without_cover) > 0:
        # at least we have album id, let's see if we have cover art
        # choose one randomly
        sel_album_without: ArtistAlbumCoverArt = secrets.choice(album_cover_art_list_without_cover)
        return sel_album_without.cover_art
    # if there is precaching, we are finished here.
    if config.get_config_param_as_bool(constants.ConfigParam.PRELOAD_SONGS):
        return None
    # load artist
    if verbose:
        msgproc.log(f"get_album_cover_art_by_artist_id loading artist by id [{artist_id}] ...")
    artist: Artist = subsonic_util.try_get_artist(artist_id=artist_id)
    artist_cover_art: str = subsonic_util.get_artist_cover_art(artist=artist) if artist else None
    if artist_cover_art:
        # use this!
        if verbose:
            msgproc.log(f"get_album_cover_art_by_artist_id using artist coverArt for artist_id [{artist_id}]")
        return artist_cover_art
    album_list: list[Album] = artist.getAlbumList() if artist else []
    select_album_list: list[Album] = album_list
    # try to avoid "appearances" first
    as_main_artist_album_list: list[Album] = subsonic_util.get_artist_albums_as_appears_on(
        artist_id=artist_id,
        album_list=album_list,
        opposite=True)
    if verbose:
        msgproc.log(f"get_album_cover_art_by_artist_id select_album_list for [{artist_id}] "
                    f"[{artist.getName() if artist else None}] "
                    f"full length [{len(album_list) if album_list else 0}] "
                    f"as main artist only [{len(as_main_artist_album_list)}]")
    if len(as_main_artist_album_list) > 0:
        select_album_list = as_main_artist_album_list
    subsonic_util.sort_albums_by_date(select_album_list)
    select_album: Album = __get_first_album_with_cover_art_from_album_list(select_album_list)
    if verbose:
        msgproc.log(f"get_album_cover_art_by_artist_id for artist_id [{artist_id}] "
                    f"select_album [{select_album.getId() if select_album else 'None'}]")
    if select_album:
        cache_actions.on_album(album=select_album)
    return select_album.getCoverArt() if select_album else None


def get_album_cover_art_url_by_artist_id(artist_id: str, skip_artist_metadata_cache: bool = False) -> str:
    cover_art: str = get_album_cover_art_by_artist_id(
        artist_id=artist_id,
        skip_artist_metadata_cache=skip_artist_metadata_cache)
    return subsonic_util.build_cover_art_url(item_id=cover_art) if cover_art else None


def __get_first_album_with_cover_art_from_album_list(album_list: list[Album]) -> str:
    album: Album
    for album in album_list if album_list else []:
        if album.getCoverArt():
            if config.get_verbose_logging():
                msgproc.log(f"__get_first_album_with_cover_art_from_album_list using album [{album.getId()}] "
                            f"coverArt [{album.getCoverArt()}]")
            return album
    # none found
    return None


def __get_first_cover_art_from_album_list(album_list: list[Album]) -> str:
    album: Album = __get_first_album_with_cover_art_from_album_list(album_list=album_list)
    return album.getCoverArt() if album else None


def favourite_albums_art_retriever(tag_to_entry_context: TagToEntryContext = None) -> RetrievedArt:
    fav: RetrievedArt = __favourite_albums_art_retriever()
    return fav if fav else random_albums_art_retriever()


def __favourite_albums_art_retriever() -> RetrievedArt:
    response: Response[Starred] = request_cache.get_starred()
    if not response.isOk():
        return None
    album_list: list[Album] = response.getObj().getAlbums()
    return __get_cover_art_from_album_list(album_list=album_list, random=True)


def favourite_artist_art_retriever(tag_to_entry_context: TagToEntryContext = None) -> RetrievedArt:
    fav: RetrievedArt = _favourite_artist_art_retriever()
    msgproc.log(f"favourite_artist_art_retriever fav found: [{'yes' if fav else 'no'}]")
    return fav if fav is not None else random_albums_art_retriever()


def _favourite_artist_art_retriever() -> RetrievedArt:
    verbose: bool = config.get_verbose_logging()
    response: Response[Starred] = request_cache.get_starred()
    if not response.isOk():
        msgproc.log("_favourite_artist_art_retriever no starred artists")
        return None
    artist_list: list[Artist] = response.getObj().getArtists()
    with_id: list[Artist] = list(filter(lambda a: a.getId() is not None, artist_list if artist_list else []))
    select_artist: Artist = secrets.choice(with_id) if with_id and len(with_id) > 0 else None
    msgproc.log(f"_favourite_artist_art_retriever select_artist is set: {'yes' if select_artist else 'no'}")
    msgproc.log("_favourite_artist_art_retriever selected artist "
                f"[{select_artist.getId() if select_artist else None}] "
                f"[{select_artist.getName() if select_artist else None}]")
    if not select_artist:
        return None
    artist_cover_art: str = subsonic_util.get_artist_cover_art(artist=select_artist)
    art_url: str = None
    if verbose:
        msgproc.log("_favourite_artist_art_retriever got cover art from artist "
                    f"[{select_artist.getId()}] [{select_artist.getName()}]: "
                    f"[{'yes' if artist_cover_art else 'no'}] "
                    f"-> CoverArt: [{artist_cover_art}]")
    if artist_cover_art:
        msgproc.log(f"_favourite_artist_art_retriever [{select_artist.getId()}] [{select_artist.getName()}] "
                    f"using artist information (best option)")
        art_url = subsonic_util.build_cover_art_url(item_id=artist_cover_art)
        return RetrievedArt(art_url=art_url)
    # look in cache first
    select_artist_metadata: persistence.ArtistMetadata = persistence.get_artist_metadata(artist_id=select_artist.getId())
    if select_artist_metadata and select_artist_metadata.artist_cover_art:
        msgproc.log(f"_favourite_artist_art_retriever [{select_artist.getId()}] [{select_artist.getName()}] "
                    f"using artist metadata (second best option)")
        art_url = subsonic_util.build_cover_art_url(item_id=select_artist_metadata.artist_cover_art)
        return RetrievedArt(art_url=art_url)
    # last chance, load artist, get albums, then select one
    msgproc.log(f"_favourite_artist_art_retriever [{select_artist.getId()}] [{select_artist.getName()}] "
                f"loading artist, worst option")
    art_url = get_album_cover_art_url_by_artist_id(artist_id=select_artist.getId())
    return RetrievedArt(art_url=art_url)


def favourite_song_retriever(tag_to_entry_context: TagToEntryContext = None) -> RetrievedArt:
    fav: RetrievedArt = _favourite_song_retriever()
    return fav if fav else random_albums_art_retriever()


def _favourite_song_retriever() -> RetrievedArt:
    return _art_for_favourite_song(random=True)


def _art_for_favourite_song(random: bool = False) -> RetrievedArt:
    response: Response[Starred] = request_cache.get_starred()
    if not response.isOk():
        return None
    song_list: list[Song] = response.getObj().getSongs()
    if not song_list or len(song_list) == 0:
        return None
    select_song: Song = secrets.choice(song_list) if random else song_list[0]
    art_url: str = (subsonic_util.build_cover_art_url(item_id=select_song.getCoverArt())
                    if select_song
                    else None)
    return RetrievedArt(art_url=art_url)


def playlists_art_retriever(tag_to_entry_context: TagToEntryContext = None) -> RetrievedArt:
    response: Response[Playlists] = request_cache.get_playlists()
    if not response.isOk():
        msgproc.log("playlists_art_retriever - cannot get playlists")
        return None
    playlist_list: list[Playlist] = response.getObj().getPlaylists()
    if not playlist_list or len(playlist_list) == 0:
        msgproc.log("playlists_art_retriever - no playlists")
        return None
    select: Playlist = secrets.choice(playlist_list)
    if not select:
        msgproc.log("playlists_art_retriever - no playlist selected")
        return None
    art_url: str = (subsonic_util.build_cover_art_url(item_id=select.getCoverArt())
                    if select and select.getCoverArt()
                    else None)
    return RetrievedArt(art_url=art_url)


def execute_art_retriever(
        tag: TagType,
        context: TagToEntryContext = None) -> RetrievedArt:
    tagname: str = tag.tag_name
    if tagname in __tag_art_retriever_dict:
        try:
            return __tag_art_retriever_dict[tagname](context)
        except Exception as ex:
            msgproc.log(f"Cannot retrieve art for tag [{tagname}] [{type(ex)}] [{ex}]")


__tag_art_retriever_dict: dict[str, Callable[[], RetrievedArt]] = {
    TagType.ALBUMS.tag_name: group_albums_art_retriever,
    TagType.ARTIST_ROLES.tag_name: group_artists_art_retriever,
    TagType.ARTISTS.tag_name: group_artists_art_retriever,
    TagType.SONGS.tag_name: group_songs_art_retriever,
    TagType.RECENTLY_ADDED_ALBUMS.tag_name: recently_added_albums_art_retriever,
    TagType.ALPHABETICAL_BY_NAME_ALBUMS.tag_name: random_albums_art_retriever,
    TagType.ALPHABETICAL_BY_ARTIST_ALBUMS.tag_name: random_albums_art_retriever,
    TagType.NEWEST_ALBUMS.tag_name: newest_albums_art_retriever,
    TagType.OLDEST_ALBUMS.tag_name: random_albums_art_retriever,
    TagType.RECENTLY_PLAYED_ALBUMS.tag_name: recently_played_albums_art_retriever,
    TagType.HIGHEST_RATED_ALBUMS.tag_name: highest_rated_albums_art_retriever,
    TagType.FAVORITE_ALBUMS.tag_name: favourite_albums_art_retriever,
    TagType.MOST_PLAYED_ALBUMS.tag_name: most_played_albums_art_retriever,
    TagType.RANDOM.tag_name: random_albums_art_retriever,
    TagType.ALBUM_BROWSER.tag_name: random_albums_art_retriever,
    TagType.RANDOM_SONGS_LIST.tag_name: random_albums_art_retriever,
    TagType.GENRES.tag_name: random_albums_art_retriever,
    TagType.ALL_ARTISTS.tag_name: random_albums_art_retriever,
    TagType.ALL_ARTISTS_UNSORTED.tag_name: random_albums_art_retriever,
    TagType.ALL_ARTISTS_INDEXED.tag_name: random_albums_art_retriever,
    TagType.FAVORITE_ARTISTS.tag_name: favourite_artist_art_retriever,
    TagType.PLAYLISTS.tag_name: playlists_art_retriever,
    TagType.FAVORITE_SONGS.tag_name: favourite_song_retriever,
    TagType.FAVORITE_SONGS_LIST.tag_name: favourite_song_retriever,
    TagType.RECENTLY_PLAYED_SONGS.tag_name: recently_played_albums_art_retriever
}



# Copyright (C) 2023,2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from item_identifier_key import ItemIdentifierKey
import copy
import random
import string
import constants
import config


class ItemIdentifier:

    def randomword(self, length) -> str:
        letters: str = string.ascii_lowercase
        return ''.join(random.choice(letters) for i in range(length))

    @classmethod
    def from_dict(cls, id_dict: dict[str, any]):
        thing_name: str = ItemIdentifier.__check_mandatory(id_dict, ItemIdentifierKey.THING_NAME)
        thing_value: any = ItemIdentifier.__check_mandatory(id_dict, ItemIdentifierKey.THING_VALUE)
        id: ItemIdentifier = cls(thing_name, thing_value)
        for k, v in id_dict.items():
            if not id.__has_name(k):
                id.__set(k, v)
        return id

    def __check_mandatory(id_dict: dict[str, any], id_key: ItemIdentifierKey) -> any:
        if id_key.identifier_name not in id_dict:
            raise Exception("Mandatory [{id_key.identifier_name}] missing")
        return id_dict[id_key.identifier_name]

    def __init__(self, name: str, value: any):
        self.__dict: dict[str, any] = {}
        if not name:
            raise Exception("name cannot be empty")
        if not value:
            raise Exception("value cannot be empty")
        self.set(ItemIdentifierKey.THING_NAME, name)
        self.set(ItemIdentifierKey.THING_VALUE, value)
        if config.get_config_param_as_bool(constants.ConfigParam.ENABLE_RANDOM_ID):
            random_value: str = self.randomword(6)
            self.set(ItemIdentifierKey.RANDOM_VALUE, random_value)

    def getDictionary(self):
        return copy.deepcopy(self.__dict)

    def has(self, key: ItemIdentifierKey):
        return self.__has_name(key.identifier_name)

    def __has_name(self, name: str):
        return name in self.__dict

    def get(self, key: ItemIdentifierKey, defaultValue: any = None):
        return self.__get(key.identifier_name, defaultValue)

    def __get(self, key_name: str, defaultValue: any = None):
        return self.__dict[key_name] if key_name in self.__dict else defaultValue

    def set(self, key: ItemIdentifierKey, value):
        if not self.__valid_key(key):
            raise Exception(f"Key {key.identifier_name if key else None} already set")
        self.__set(key.identifier_name, value)

    def __set(self, key_name: str, value):
        if not self.__valid_key_name(key_name):
            raise Exception(f"Key {key_name} already set")
        self.__dict[key_name] = value

    def __valid_key(self, key: ItemIdentifierKey) -> bool:
        return key and self.__valid_key_name(key.identifier_name)

    def __valid_key_name(self, key_name: str) -> bool:
        return key_name and key_name not in self.__dict



# Copyright (C) 2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from column_name import ColumnName
from typing import Any
from typing import Optional
from typing import Protocol


# We still use a small Protocol just so the linter knows the Enum
# members have a .column_name.value attribute.
class MetadataMember(Protocol):
    @property
    def column_name(self) -> ColumnName: ...


class Metadata:
    def __init__(self):
        self.__data: dict[Any, Any] = {}

    def _set(self, key: Any, value: Any) -> None:
        """Internal helper to update the data dictionary."""
        if value is None:
            return
        if isinstance(value, str) and len(value) == 0:
            return
        self.__data[key] = value

    def _get(self, key: Any) -> Optional[Any]:
        """Internal helper to retrieve values from the data dictionary."""
        return self.__data.get(key)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.__data})"



# Copyright (C) 2023,2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum
import idgenerator
import constants
import config


base_chars: str = "abcdefghijklmnopqrtuvwxyz0123456789"
current_id: int = None


class _TagTypeData:

    def __init__(self, tag_title: str, query_type: str = None):
        global current_id
        if current_id is None:
            current_id = 0
        else:
            current_id += 1
        self.__tag_name: str = idgenerator.number_to_base_decoded(n=current_id, base_chars=base_chars)
        self.__tag_title: str = tag_title
        self.__query_type: str = query_type

    @property
    def tag_name(self) -> str:
        return self.__tag_name

    @property
    def tag_title(self) -> str:
        return self.__tag_title

    @property
    def query_type(self) -> str:
        return self.__query_type


class TagType(Enum):

    ALBUMS = _TagTypeData("Albums", None)
    RECENTLY_ADDED_ALBUMS = _TagTypeData("Recently Added Albums", query_type="newest")
    RECENTLY_PLAYED_ALBUMS = _TagTypeData("Recently Played Albums", query_type="recent")
    HIGHEST_RATED_ALBUMS = _TagTypeData("Highest Rated Albums", query_type="highest")
    FAVORITE_ALBUMS = _TagTypeData("Favorite Albums", query_type="starred")
    MOST_PLAYED_ALBUMS = _TagTypeData("Most Played Albums", query_type="frequent")
    NEWEST_ALBUMS = _TagTypeData("Newest Albums", query_type="byYear")
    OLDEST_ALBUMS = _TagTypeData("Oldest Albums", query_type="byYear")
    ALPHABETICAL_BY_NAME_ALBUMS = _TagTypeData("Alphabetical By Name", query_type="alphabeticalByName")
    ALPHABETICAL_BY_ARTIST_ALBUMS = _TagTypeData("Alphabetical By Artist", query_type="alphabeticalByArtist")
    RANDOM = _TagTypeData("Random Albums", query_type="random")
    ARTIST_ROLES = _TagTypeData("Artist Roles")
    ARTISTS = _TagTypeData("Artists")
    ALL_ARTISTS = _TagTypeData("All Artists (Sorted) (slow!)")
    ALL_ARTISTS_INDEXED = _TagTypeData("All Artists (By Initial) (slow!)")
    ALL_ARTISTS_UNSORTED = _TagTypeData("All Artists")
    ALL_ALBUM_ARTISTS_UNSORTED = _TagTypeData("Album Artists")
    ALL_COMPOSERS_UNSORTED = _TagTypeData("Composers")
    ALL_CONDUCTORS_UNSORTED = _TagTypeData("Conductors")
    FAVORITE_ARTISTS = _TagTypeData("Favorite Artists", "starred")
    SONGS = _TagTypeData("Songs")
    RANDOM_SONGS_LIST = _TagTypeData("Random Songs")
    FAVORITE_SONGS = _TagTypeData("Favorite Songs (Paginated)")
    FAVORITE_SONGS_LIST = _TagTypeData("Favorite Songs")
    RECENTLY_PLAYED_SONGS = _TagTypeData("Recently played Songs")
    GENRES = _TagTypeData("Genres")
    PLAYLISTS = _TagTypeData("Playlists")
    INTERNET_RADIOS = _TagTypeData("Internet Radios")
    ALBUM_BROWSER = _TagTypeData("Album Browser")

    @property
    def tag_name(self) -> str:
        if config.get_config_param_as_bool(constants.ConfigParam.MINIMIZE_IDENTIFIER_LENGTH):
            return self.name
        else:
            return self.value.tag_name

    @property
    def tag_title(self) -> str:
        return self.value.tag_title

    @property
    def query_type(self) -> str:
        return self.value.query_type


def get_tag_type_by_name(tag_name: str) -> TagType:
    tag: TagType
    for tag in TagType:
        if tag_name == tag.tag_name:
            return tag
    raise Exception(f"get_tag_type_by_name with {tag_name} NOT found")



# Copyright (C) 2023,2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum
import idgenerator
import config
import constants

base_chars: str = "abcdefghijklmnopqrtuvwxyz0123456789"
current_id: int = None


class _ItemIdentifierKey:

    def __init__(self):
        global current_id
        if current_id is None:
            current_id = 0
        else:
            current_id += 1
        self.__element_name: str = idgenerator.number_to_base_decoded(n=current_id, base_chars=base_chars)

    @property
    def identifier_name(self) -> str:
        return self.__element_name


class ItemIdentifierKey(Enum):
    THING_NAME = _ItemIdentifierKey()
    THING_VALUE = _ItemIdentifierKey()
    GENRE_NAME = _ItemIdentifierKey()
    PAGE_NUMBER = _ItemIdentifierKey()
    ALBUM_ID = _ItemIdentifierKey()
    OFFSET = _ItemIdentifierKey()
    TAG_TYPE = _ItemIdentifierKey()
    ALBUM_VERSION_PATH_BASE64 = _ItemIdentifierKey()
    RADIO_NAME = _ItemIdentifierKey()
    SONG_AS_NAVIGABLE_ENTRY = _ItemIdentifierKey()
    RANDOM_VALUE = _ItemIdentifierKey()
    SKIP_ARTIST_ID = _ItemIdentifierKey()
    ALBUM_RELEASE_TYPE = _ItemIdentifierKey()
    ALBUM_ID_REF_FOR_ARTIST = _ItemIdentifierKey()
    ALBUM_DISC_NUMBERS = _ItemIdentifierKey()
    ALBUM_IGNORE_DISCNUMBERS = _ItemIdentifierKey()
    ARTIST_ROLE = _ItemIdentifierKey()
    ALBUM_BROWSE_SELECTION_LIST = _ItemIdentifierKey()
    ALBUM_BROWSE_FILTER_KEY = _ItemIdentifierKey()

    @property
    def identifier_name(self) -> str:
        if config.get_config_param_as_bool(constants.ConfigParam.MINIMIZE_IDENTIFIER_LENGTH):
            return self.name
        else:
            return self.value.element_name


# duplicate check
name_checker_set: set[str] = set()
for v in ItemIdentifierKey:
    if v.identifier_name in name_checker_set:
        raise Exception(f"Duplicated name [{v.identifier_name}]")
    name_checker_set.add(v.identifier_name)



#!/usr/bin/env python3

import sys
import os
import getopt

thisdir=os.path.dirname(__file__)
sys.path.append(os.path.join(thisdir, "../pycommon"))
from upmplgutils import *
from conftree import ConfSimple

from api.bundle import Bundle

import socket
def get_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        # doesn't even have to be reachable
        s.connect(('10.254.254.254', 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP

## We would like to check the content of the cache config file, but this is complicated because
## it depends on upmpdcli being started as root (cache in /var), or as a regular user (cache in
## ~/.cache). So the following is not used for now.
def check_success():
    input("Please type <CR> when you get the 'successful' page in the browser")
    
    config = ConfSimple(os.path.join(getcachedir("qobuz"), "config"))
    auth_token = config.get("user_auth_token")
    user_id = config.get("user_id")
    if auth_token and user_id:
        print("Successfully got user id and authorization token")
    else:
        print("Seems to have failed")
    

def perform_oauth_login(host, port):
    bundle = Bundle()
    app_id = str(bundle.get_app_id())

    oauth_url_base = (
        f"https://www.qobuz.com/signin/oauth"
        f"?ext_app_id={app_id}"
        
    )
    oauth_url_local = oauth_url_base + f"&redirect_url=http://localhost:{port}/qobuz/oauth/"
    oauth_url_net = oauth_url_base + f"&redirect_url=http://{host}:{port}/qobuz/oauth/"

    print("This script must run on the same machine as upmpdcli")
    print("Open one of the following URLs in your browser to authenticate with Qobuz:")
    print()
    print("- If upmpdcli and the script run on the same machine as the WEB browser, use:")
    print(f"{oauth_url_local}")
    print()
    print("- If the upmpdcli and the script run on a different machine than the browser, use:")
    print(f"{oauth_url_net}")


def msg(s):
    print(f"{s}", file=sys.stderr)
    
###### Main
configfile = "/etc/upmpdcli.conf"
opts = getopt.getopt(sys.argv[1:], "c:")
for opt, arg in opts[0]:
    if opt == "-c":
        configfile = arg

if not os.path.exists(configfile):
    msg("Upmpdcli configuration file not found in the default location (/etc/upmpdcli.conf)")
    msg("Please use the -c option to set the configuration file location")
    sys.exit(1)
os.environ["UPMPD_CONFIG"] = configfile

host = getOptionValue("plgmicrohttphost", get_ip())
port = getOptionValue("plgmicrohttpport", 49149)

perform_oauth_login(host, port)



# Copyright (C) 2024,2025,2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cache_type
import cache_manager_provider
import subsonic_util
import config
import persistence
import album_util
import time
import metadata_converter
from subsonic_connector.album import Album
from subsonic_connector.song import Song
from artist_from_album import ArtistFromAlbum
from song_data_structures import SongArtistType
from song_data_structures import SongArtist
from song_data_structures import SongContributor
from table_name import TableName
from metadata_model import SongMetadataModel
from msgproc_provider import msgproc
from typing import Any
import sqlite3

def delete_key(cache_type: cache_type.CacheType, key: str) -> bool:
    return cache_manager_provider.get().delete_cached_element(cache_type.cache_name, key)


def on_album(album: Album):
    start: float = time.time()
    connection: sqlite3.Connection = persistence.get_working_connection()
    try:
        __on_album(album=album, connection=connection)
    except Exception as ex:
        msgproc.log(f"on_album failed due to [{type(ex)}] [{ex}]")
        connection.close()
    elapsed: float = time.time() - start
    if config.get_verbose_logging():
        msgproc.log(f"on_album for album_id [{album.getId()}] executed in [{elapsed:.3f}]")


def __on_album(album: Album, connection: sqlite3.Connection = None):
    if not album or not album.getId():
        # nothing to do
        return
    verbose: bool = config.get_verbose_logging()
    the_connection: sqlite3.Connection = persistence.get_working_connection(connection)
    album_quality_badge: str = None
    track_quality_summary: str = None
    album_path_joined: str = None
    song_list: list[Song] = album.getSongs()
    song_count: int = len(song_list) if song_list else 0
    if len(song_list) if song_list else 0:
        # we can calculate qualities and path
        album_quality_badge = subsonic_util.calc_song_list_quality_badge(song_list=song_list)
        track_quality_summary = subsonic_util.calc_song_quality_summary(song_list=song_list)
        album_path_joined = album_util.get_album_path_list_joined(song_list=song_list)
        if verbose:
            msgproc.log(f"__on_album [{album.getId()}] -> "
                        f"badge [{album_quality_badge}] "
                        f"summary [{track_quality_summary}] "
                        f"path [{album_path_joined}]")
    # delete orphaned tracks
    # msgproc.log("delete_song_list_not_in")
    persistence.delete_song_list_not_in(
        album_id=album.getId(),
        song_list=list(map(lambda x: x.getId(), song_list)),
        connection=the_connection,
        do_commit=False)
    # the_connection.commit()
    # save metadata
    # msgproc.log("save_album_metadata")
    persistence.save_album_metadata(
        album_metadata=metadata_converter.build_album_metadata(
            album=album,
            quality_badge=album_quality_badge,
            song_quality_summary=track_quality_summary,
            album_path=album_path_joined),
        context="__on_album",
        connection=the_connection,
        do_commit=False)
    # save properties
    album_properties: dict[str, list[Any]] = subsonic_util.build_album_properties(album=album)
    # msgproc.log("save_album_properties")
    persistence.save_album_properties(
        album_id=album.getId(),
        properties=album_properties,
        connection=the_connection,
        do_commit=False)
    # save other data
    album_artists: list[ArtistFromAlbum] = subsonic_util.get_artists_from_album(album=album)
    # msgproc.log("update_album_artists")
    persistence.update_album_artists(
        album_id=album.getId(),
        album_artists=album_artists,
        connection=the_connection,
        do_commit=False)
    # msgproc.log("update_album_discs")
    persistence.update_album_discs(
        album_id=album.getId(),
        album_discs=subsonic_util.get_disc_titles_from_album(album=album),
        connection=the_connection,
        do_commit=False)
    # msgproc.log("update_album_genres")
    persistence.update_album_genres(
        album_id=album.getId(),
        album_genres=subsonic_util.get_genres_from_album(album=album),
        connection=the_connection,
        do_commit=False)
    # msgproc.log("update_album_record_labels")
    persistence.update_album_record_labels(
        album_id=album.getId(),
        album_record_labels=subsonic_util.get_album_record_label_names(album=album),
        connection=the_connection,
        do_commit=False)
    # msgproc.log("update_album_moods")
    persistence.update_album_moods(
        album_id=album.getId(),
        album_moods=subsonic_util.get_album_moods(album=album),
        connection=the_connection,
        do_commit=False)
    # msgproc.log("update_album_release_types")
    persistence.update_album_release_types(
        album_id=album.getId(),
        album_release_types=subsonic_util.get_album_release_types(album=album).types,
        connection=the_connection,
        do_commit=False)
    if verbose:
        msgproc.log("__on_album saving songs ...")
    song: Song
    for song in song_list if song_list else []:
        # msgproc.log("__on_album_song")
        __on_album_song(
            album=album,
            song=song,
            connection=the_connection,
            do_commit=False)
    # purge removed songs
    if song_count > 0:
        persistence.delete_by_parent_id_and_id_in_list(
            table_name=TableName.SONG_METADATA_V1,
            parent_id_column_name=SongMetadataModel.SONG_ALBUM_ID.column_name,
            parent_id=album.getId(),
            id_column_name=SongMetadataModel.SONG_ID.column_name,
            id_list=list(map(lambda x: x.getId(), song_list)),
            in_mode=persistence.InMode.NOT_IN,
            connection=the_connection,
            do_commit=False)
    persistence.commit(connection=the_connection)
    if connection is None:
        the_connection.close()


def __on_album_song(
        album: Album,
        song: Song,
        connection: sqlite3.Connection,
        do_commit: bool = True):
    verbose: bool = config.get_verbose_logging()
    the_connection: sqlite3.Connection = persistence.get_working_connection(provided=connection)
    if verbose:
        msgproc.log(f"__on_album saving songs song [{song.getId()}] ...")
    # msgproc.log("__on_album_song save_song_metadata")
    persistence.save_song_metadata(
        song_metadata=metadata_converter.build_song_metadata(song=song),
        context="__on_album_song",
        connection=connection,
        do_commit=do_commit)
    # song album artists
    song_album_artist_list: list[SongArtist] = subsonic_util.get_song_artists_by_type(
        song=song,
        song_artist_type=SongArtistType.SONG_ALBUM_ARTIST)
    # msgproc.log("__on_album_song save_song_album_artist_list")
    persistence.save_song_album_artist_list(
        song_id=song.getId(),
        album_id=album.getId(),
        song_album_artist_list=song_album_artist_list,
        connection=the_connection,
        do_commit=do_commit)
    if verbose:
        msgproc.log(f"__on_album [{album.getId()}] song [{song.getId()}] album_artists [{len(song_album_artist_list)}]")
    # song artists
    song_artist_list: list[SongArtist] = subsonic_util.get_song_artists_by_type(
        song=song,
        song_artist_type=SongArtistType.SONG_ARTIST)
    # msgproc.log("__on_album_song save_song_artist_list")
    persistence.save_song_artist_list(
        song_id=song.getId(),
        album_id=album.getId(),
        song_artist_list=song_artist_list,
        connection=the_connection,
        do_commit=do_commit)
    if verbose:
        msgproc.log(f"__on_album [{album.getId()}] song [{song.getId()}] artists [{len(song_artist_list)}]")
    # contributors
    song_contributor_list: list[SongContributor] = subsonic_util.get_song_contributors(song=song)
    # msgproc.log("__on_album_song save_song_contributor_list")
    persistence.save_song_contributor_list(
        song_id=song.getId(),
        album_id=album.getId(),
        song_contributor_list=song_contributor_list,
        connection=the_connection,
        do_commit=do_commit)
    if verbose:
        msgproc.log(f"__on_album [{album.getId()}] song [{song.getId()}] contributors [{len(song_contributor_list)}]")



# Copyright (C) 2023,2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum


class TileType(Enum):
    TAG = 1, "tag",
    CATEGORY = 2, "category"
    PAGE_LINK = 3, "pagelink"

    def __init__(
            self,
            type_type_num: int,
            tile_type_name: str):
        self._type_type_num: int = type_type_num
        self._tile_type_name: str = tile_type_name

    @property
    def tile_type_num(self) -> int:
        """I'm the 'tile_type_num' property."""
        return self._type_type_num

    @property
    def tile_type_name(self) -> str:
        """I'm the 'tile_type_name' property."""
        return self._tile_type_name


def get_tile_type_by_name(tile_type_name: str) -> TileType:
    # msgproc.log(f"get_tidal_tag_type_by_name with {tag_name}")
    for _, member in TileType.__members__.items():
        if tile_type_name == member.tile_type_name:
            return member
    raise Exception(f"get_tile_type_by_name with {tile_type_name} NOT found")


# duplicate check
name_checker_set: set[str] = set()
id_checker_set: set[int] = set()
for v in TileType:
    if v.tile_type_name in name_checker_set:
        raise Exception(f"Duplicated name [{v.tile_type_name}]")
    if v.tile_type_num in id_checker_set:
        raise Exception(f"Duplicated id [{v.tile_type_num}]")
    name_checker_set.add(v.tile_type_name)
    id_checker_set.add(v.tile_type_num)



##########################################################################
# Copyright 2009 Carlos Ribeiro
#
# This file is part of Radio Tray
#
# Radio Tray is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 1 of the License, or
# (at your option) any later version.
#
# Radio Tray is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radio Tray.  If not, see <http://www.gnu.org/licenses/>.
#
##########################################################################
import sys
from urllib.request import Request as UrlRequest
from urllib.request import urlopen as urlUrlopen
from urllib.parse import urljoin as urlJoin

import logging    
from common import USER_AGENT

class M3uPlaylistDecoder:
    def __init__(self):
        self.log = logging.getLogger('upmpdcli')
        #self.log.setLevel(logging.DEBUG)


    def isStreamValid(self, contentType, firstBytes):
        contentType = contentType.lower()

        # Apple live streaming special case: we do not try to process these, there is no permanent
        # media url to extract (only an ever changing list of short segments). We reject them and
        # let mpd deal with the playlist itself (which it does well, even if I could not find out by
        # looking at the code how it worked).
        if 'audio/mpegurl' in contentType or 'application/vnd.apple.mpegurl' in contentType:
            lines = firstBytes.splitlines()
            if not len(lines) or lines[0].startswith(b"#EXTM3U"):
                return False
        
        if 'audio/mpegurl' in contentType or 'audio/x-mpegurl' in contentType \
           or 'application/x-mpegurl' in contentType or \
           'application/vnd.apple.mpegurl' in contentType:
            # we tested for EXTM3U above, so this must be a regular playlist
            self.log.debug('Stream is readable by M3U Playlist Decoder')
            return True
        else:
            lines = firstBytes.splitlines()
            for line in lines:
                if line.startswith(b"http://"):
                    return True
        return False


    def extractPlaylist(self,  url):
        self.log.info('M3u: downloading playlist...')
        req = UrlRequest(url)
        req.add_header('User-Agent', USER_AGENT)
        f = urlUrlopen(req)
        str = f.read()
        f.close()

        self.log.info('M3U: playlist downloaded, decoding... ')

        lines = str.splitlines()
        playlist = []

        for line in lines:
            if line.startswith(b"#") == False and len(line) > 0:
                if line.startswith(b"http"):
                    playlist.append(line)
                else:
                    playlist.append(urlJoin(url.encode('utf-8'), line))

        return playlist



# Copyright (C) 2023,2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from enum import Enum
# from msgproc_provider import msgproc


class LavfMatchMode(Enum):
    EQ = 1  # exact only
    GE = 2  # exact or better


class LavfConstant(Enum):
    PREFIX = "Lavf/"
    INITIAL_SPLITTER = "/"
    VERSION_SPLITTER = "."


def match(to_match: str, pattern: str, match_mode: LavfMatchMode = LavfMatchMode.EQ) -> bool:
    # msgproc.log(f"Matching on to_match: [{to_match}] pattern: [{pattern}] mode: [{match_mode}]")
    # sanity checks.
    if (not to_match or
            not pattern or
            not to_match.startswith(LavfConstant.PREFIX.value) or
            not pattern.startswith(LavfConstant.PREFIX.value)):
        # msgproc.log(f"Invalid to_match [{to_match}] or pattern [{pattern}]")
        return False
    # can do the split?
    to_match_splitted: list[str] = to_match.split(LavfConstant.INITIAL_SPLITTER.value)
    pattern_splitted: list[str] = pattern.split(LavfConstant.INITIAL_SPLITTER.value)
    if not to_match_splitted or len(to_match_splitted) != 2:
        return False
    if not pattern_splitted or len(pattern_splitted) != 2:
        # msgproc.log(f"Cannot split, len is [{len(pattern_splitted) if pattern_splitted else 0}]")
        return False
    # split complete, check versions.
    to_match_version: str = to_match_splitted[1]
    pattern_version: str = pattern_splitted[1]
    # msgproc.log(f"to_match_version [{to_match_version}] pattern_version [{pattern_version}]")
    same: bool = to_match_version == pattern_version
    if same:
        return True
    if not same and match_mode == LavfMatchMode.EQ:
        return False
    # compare the versions.
    to_match_v_array: list[str] = to_match_version.split(LavfConstant.VERSION_SPLITTER.value)
    pattern_v_array: list[str] = pattern_version.split(LavfConstant.VERSION_SPLITTER.value)
    for i in range(0, min(len(to_match_v_array), len(pattern_v_array))):
        t: int = _to_int(to_match_v_array[i])
        p: int = _to_int(pattern_v_array[i])
        if t is not None and p is not None:
            if t > p:
                return True
            elif t < p:
                return False
        else:
            return False
    # if we are here, they are the same version if the length is the same
    if len(to_match_v_array) >= len(pattern_v_array):
        return True
    return False


def _to_int(v: str) -> int:
    try:
        return int(v)
    except Exception:
        return None



# Copyright (C) 2023,2024,2025,2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class RadioStationEntry:

    def __init__(
            self,
            codec: str,
            url: str,
            title: str,
            mimetype: str,
            bit_depth: int,
            sampling_rate: int,
            channel_count: int = 2):
        self.__title: str = title
        self.__url: str = url
        self.__codec: str = codec
        self.__mimetype: str = mimetype
        self.__bit_depth: int = bit_depth
        self.__sampling_rate: int = sampling_rate
        self.__channel_count: int = channel_count

    @property
    def title(self) -> str:
        return self.__title

    @property
    def url(self) -> str:
        return self.__url
    
    @property
    def codec(self) -> str:
        return self.__codec

    @property
    def mimetype(self) -> str:
        return self.__mimetype

    @property
    def bit_depth(self) -> int:
        return self.__bit_depth

    @property
    def sampling_rate(self) -> int:
        return self.__sampling_rate

    @property
    def channel_count(self) -> int:
        return self.__channel_count



##########################################################################
# Copyright 2009 Carlos Ribeiro
#
# This file is part of Radio Tray
#
# Radio Tray is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 1 of the License, or
# (at your option) any later version.
#
# Radio Tray is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radio Tray.  If not, see <http://www.gnu.org/licenses/>.
#
##########################################################################

class UrlInfo(object):

    def __init__(self, url, playlist, contentType, decoder = None):
        self.url = url
        self.playlist = playlist
        self.contentType = contentType
        self.decoder = decoder

    def isPlaylist(self):
        return self.playlist

    def getContentType(self):
        return self.contentType

    def getDecoder(self):
        return self.decoder

    def getUrl(self):
        return self.url



# Copyright (C) 2023,2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class CacheManager:

    def __init__(self):
        self.__caches: dict[str, any] = {}

    def __get_element_cache(self, cache_name: str) -> dict[str, any]:
        if cache_name in self.__caches:
            return self.__caches[cache_name]
        cache: dict[str, any] = {}
        self.__caches[cache_name] = cache
        return cache

    def cache_element_value(self, cache_name: str, key: str, value: any):
        cache: dict[str, any] = self.__get_element_cache(cache_name)
        cache[key] = value

    def cache_element_multi_value(self, cache_name: str, key: str, value: any):
        cache: dict[str, any] = self.__get_element_cache(cache_name)
        existing: set[any] = cache[key] if key in cache else None
        if not existing:
            existing = set()
            cache[key] = existing
        if value not in existing:
            existing.add(value)

    def is_element_cached(self, cache_name: str, key: str) -> bool:
        cache: dict[str, any] = self.__get_element_cache(cache_name)
        return key in cache if cache else False

    def get_cached_element(self, cache_name: str, key: str) -> any:
        cache: dict[str, any] = self.__get_element_cache(cache_name)
        return cache[key] if key in cache else None

    def delete_cached_element(self, cache_name: str, key: str) -> bool:
        cache: dict[str, any] = self.__get_element_cache(cache_name)
        can_delete: bool = key in cache
        if can_delete:
            del cache[key]
        return can_delete

    def get_cache_size(self, cache_name: str) -> int:
        element_cache: dict[str, any] = self.__get_element_cache(cache_name)
        return len(element_cache) if element_cache else 0



# Copyright (C) 2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from metadata import Metadata
from metadata_model import AlbumMetadataModel
from typing import Optional
from typing import Any
import datetime


class AlbumMetadata(Metadata):

    def get_value(self, album_metadata_model: AlbumMetadataModel, dflt: Any = None) -> Optional[Any]:
        v: Any = self._get(album_metadata_model.column_name)
        if v is None and dflt is not None:
            v = dflt
        return v

    def set_value(self, album_metadata_model: AlbumMetadataModel, value: Any):
        return self._set(album_metadata_model.column_name, value)

    @property
    def album_id(self) -> str:
        return self.get_value(AlbumMetadataModel.ALBUM_ID)

    @property
    def quality_badge(self) -> str:
        return self.get_value(AlbumMetadataModel.QUALITY_BADGE)

    @property
    def album_media_type(self) -> str:
        return self.get_value(AlbumMetadataModel.ALBUM_MEDIA_TYPE)

    @property
    def album_lossless_status(self) -> str:
        return self.get_value(AlbumMetadataModel.ALBUM_LOSSLESS_STATUS)

    @property
    def album_average_bitrate(self) -> int:
        return self.get_value(AlbumMetadataModel.ALBUM_AVERAGE_BITRATE)

    @property
    def album_musicbrainz_id(self) -> str:
        return self.get_value(AlbumMetadataModel.ALBUM_MB_ID)

    @property
    def album_artist_id(self) -> str:
        return self.get_value(AlbumMetadataModel.ALBUM_ARTIST_ID)

    @property
    def album_path(self) -> str:
        return self.get_value(AlbumMetadataModel.ALBUM_PATH)

    @property
    def album_artist(self) -> str:
        return self.get_value(AlbumMetadataModel.ALBUM_ARTIST)

    @property
    def album_version(self) -> str:
        return self.get_value(AlbumMetadataModel.ALBUM_VERSION)

    @property
    def album_cover_art(self) -> str:
        return self.get_value(AlbumMetadataModel.ALBUM_COVER_ART)

    @property
    def album_name(self) -> str:
        return self.get_value(AlbumMetadataModel.ALBUM_NAME)

    @property
    def album_disc_count(self) -> int:
        return self.get_value(AlbumMetadataModel.ALBUM_DISC_COUNT)

    @property
    def album_song_count(self) -> int:
        return self.get_value(AlbumMetadataModel.ALBUM_SONG_COUNT)

    @property
    def album_duration(self) -> int:
        return self.get_value(AlbumMetadataModel.ALBUM_DURATION)

    @property
    def album_created(self) -> datetime.datetime:
        return self.get_value(AlbumMetadataModel.ALBUM_CREATED)

    @property
    def album_year(self) -> int:
        return self.get_value(AlbumMetadataModel.ALBUM_YEAR)

    @property
    def album_genre(self) -> str:
        return self.get_value(AlbumMetadataModel.ALBUM_GENRE)

    @property
    def album_genre_list(self) -> str:
        return self.get_value(AlbumMetadataModel.ALBUM_GENRE_LIST)

    @property
    def album_mood_list(self) -> str:
        return self.get_value(AlbumMetadataModel.ALBUM_MOOD_LIST)

    @property
    def album_record_label_list(self) -> str:
        return self.get_value(AlbumMetadataModel.ALBUM_RECORD_LABEL_LIST)

    @property
    def album_release_type_list(self) -> str:
        return self.get_value(AlbumMetadataModel.ALBUM_RELEASE_TYPE_LIST)

    @property
    def album_user_rating(self) -> int:
        return self.get_value(AlbumMetadataModel.ALBUM_USER_RATING)

    @property
    def album_display_artist(self) -> int:
        return self.get_value(AlbumMetadataModel.ALBUM_DISPLAY_ARTIST, "")

    @property
    def album_explicit_status(self) -> int:
        return self.get_value(AlbumMetadataModel.ALBUM_EXPLICIT_STATUS)

    @property
    def album_track_quality_summary(self) -> str:
        return self.get_value(AlbumMetadataModel.ALBUM_TRACK_QUALITY_SUMMARY)

    @property
    def album_release_date_str(self) -> str:
        return self.get_value(AlbumMetadataModel.ALBUM_RELEASE_DATE_STR, "")

    @property
    def album_original_release_date_str(self) -> str:
        return self.get_value(AlbumMetadataModel.ALBUM_ORIGINAL_RELEASE_DATE_STR, "")

    @property
    def created_timestamp(self) -> datetime.datetime:
        return self.get_value(AlbumMetadataModel.CREATED_TIMESTAMP)

    @property
    def updated_timestamp(self) -> datetime.datetime:
        return self.get_value(AlbumMetadataModel.UPDATED_TIMESTAMP)
